# pylint: disable=C0103,E1101
import json
import os
import shutil
import socket
import sys
import tarfile
import tempfile
import unittest
import uuid
from ssl import SSLError
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    Iterable,
    List,
    Literal,
    Set,
    Tuple,
    Union,
)
from urllib.error import URLError
from urllib.request import urlopen

import pytest

import bioblend
from bioblend.galaxy import dataset_collections
from bioblend.galaxy.objects import (
    galaxy_instance,
    wrappers,
)
from . import test_util

bioblend.set_stream_logger("test", level="INFO")
socket.setdefaulttimeout(10.0)
SAMPLE_FN = test_util.get_abspath(os.path.join("data", "paste_columns.ga"))
SAMPLE_WF_COLL_FN = test_util.get_abspath(os.path.join("data", "paste_columns_collections.ga"))
SAMPLE_WF_PARAMETER_INPUT_FN = test_util.get_abspath(os.path.join("data", "workflow_with_parameter_input.ga"))
FOO_DATA = "foo\nbar\n"
FOO_DATA_2 = "foo2\nbar2\n"
SAMPLE_WF_DICT = {
    "deleted": False,
    "id": "9005c5112febe774",
    "inputs": {
        "571": {"label": "Input Dataset", "value": ""},
        "572": {"label": "Input Dataset", "value": ""},
    },
    "model_class": "StoredWorkflow",
    "name": "paste_columns",
    "owner": "user_foo",
    "published": False,
    "steps": {
        "571": {
            "id": 571,
            "input_steps": {},
            "tool_id": None,
            "tool_inputs": {"name": "Input Dataset"},
            "tool_version": None,
            "type": "data_input",
        },
        "572": {
            "id": 572,
            "input_steps": {},
            "tool_id": None,
            "tool_inputs": {"name": "Input Dataset"},
            "tool_version": None,
            "type": "data_input",
        },
        "573": {
            "id": 573,
            "input_steps": {
                "input1": {"source_step": 571, "step_output": "output"},
                "input2": {"source_step": 572, "step_output": "output"},
            },
            "tool_id": "Paste1",
            "tool_inputs": {
                "delimiter": '"T"',
                "input1": "null",
                "input2": "null",
            },
            "tool_version": "1.0.0",
            "type": "tool",
        },
    },
    "tags": [],
    "url": "/api/workflows/9005c5112febe774",
}
SAMPLE_INV_DICT: Dict[str, Any] = {
    "history_id": "2f94e8ae9edff68a",
    "id": "df7a1f0c02a5b08e",
    "inputs": {"0": {"id": "a7db2fac67043c7e", "src": "hda", "uuid": "7932ffe0-2340-4952-8857-dbaa50f1f46a"}},
    "model_class": "WorkflowInvocation",
    "state": "ready",
    "steps": [
        {
            "action": None,
            "id": "d413a19dec13d11e",
            "job_id": None,
            "model_class": "WorkflowInvocationStep",
            "order_index": 0,
            "state": None,
            "update_time": "2015-10-31T22:00:26",
            "workflow_step_id": "cbbbf59e8f08c98c",
            "workflow_step_label": None,
            "workflow_step_uuid": "b81250fd-3278-4e6a-b269-56a1f01ef485",
        },
        {
            "action": None,
            "id": "2f94e8ae9edff68a",
            "job_id": "e89067bb68bee7a0",
            "model_class": "WorkflowInvocationStep",
            "order_index": 1,
            "state": "new",
            "update_time": "2015-10-31T22:00:26",
            "workflow_step_id": "964b37715ec9bd22",
            "workflow_step_label": None,
            "workflow_step_uuid": "e62440b8-e911-408b-b124-e05435d3125e",
        },
    ],
    "update_time": "2015-10-31T22:00:26",
    "uuid": "c8aa2b1c-801a-11e5-a9e5-8ca98228593c",
    "workflow_id": "03501d7626bd192f",
}


def is_reachable(url: str) -> bool:
    res = None
    try:
        res = urlopen(url, timeout=5)
    except (SSLError, URLError, socket.timeout):
        return False
    if res is not None:
        res.close()
    return True


def upload_from_fs(
    lib: wrappers.Library, bnames: Iterable[str], **kwargs: Any
) -> Tuple[List[wrappers.LibraryDataset], List[str]]:
    tempdir = tempfile.mkdtemp(prefix="bioblend_test_")
    try:
        fnames = [os.path.join(tempdir, _) for _ in bnames]
        for fn in fnames:
            with open(fn, "w") as f:
                f.write(FOO_DATA)
        dss = lib.upload_from_galaxy_fs(fnames, **kwargs)
    finally:
        shutil.rmtree(tempdir)
    return dss, fnames


class MockWrapper(wrappers.Wrapper):
    BASE_ATTRS = ("a", "b")
    a: int
    b: List[int]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class TestWrapper(unittest.TestCase):
    def setUp(self):
        self.d: Dict[str, Any] = {"a": 1, "b": [2, 3], "c": {"x": 4}}
        with pytest.raises(TypeError):
            wrappers.Wrapper(self.d)
        self.w = MockWrapper(self.d)

    def test_initialize(self):
        for k in MockWrapper.BASE_ATTRS:
            assert getattr(self.w, k) == self.d[k]
        self.w.a = 222
        self.w.b[0] = 222
        assert self.w.a == 222
        assert self.w.b[0] == 222
        assert self.d["a"] == 1
        assert self.d["b"][0] == 2
        with pytest.raises(AttributeError):
            _ = self.w.foo  # type: ignore[attr-defined]
        with pytest.raises(AttributeError):
            self.w.foo = 0  # type: ignore[assignment]

    def test_taint(self):
        assert not self.w.is_modified
        self.w.a = 111  # pylint: disable=W0201
        assert self.w.is_modified

    def test_serialize(self):
        w = MockWrapper.from_json(self.w.to_json())
        assert w.wrapped == self.w.wrapped

    def test_clone(self):
        w = self.w.clone()
        assert w.wrapped == self.w.wrapped
        w.b[0] = 111
        assert self.w.b[0] == 2

    def test_kwargs(self):
        parent = MockWrapper({"a": 10})
        w = MockWrapper(self.d, parent=parent)
        assert w.parent is parent
        with pytest.raises(AttributeError):
            w.parent = 0  # type: ignore[assignment,misc]


@test_util.skip_unless_galaxy()
class GalaxyObjectsTestBase(unittest.TestCase):
    gi: galaxy_instance.GalaxyInstance

    @classmethod
    def setUpClass(cls) -> None:
        galaxy_key = os.environ["BIOBLEND_GALAXY_API_KEY"]
        galaxy_url = os.environ["BIOBLEND_GALAXY_URL"]
        cls.gi = galaxy_instance.GalaxyInstance(galaxy_url, api_key=galaxy_key)


class TestWorkflow(GalaxyObjectsTestBase):
    def setUp(self):
        self.wf = wrappers.Workflow(SAMPLE_WF_DICT)

    def test_initialize(self):
        assert self.wf.id == "9005c5112febe774"
        assert self.wf.name == "paste_columns"
        assert not self.wf.deleted
        assert self.wf.owner == "user_foo"
        assert not self.wf.published
        assert self.wf.tags == []
        assert self.wf.input_labels_to_ids == {"Input Dataset": {"571", "572"}}
        assert self.wf.tool_labels_to_ids == {"Paste1": {"573"}}
        assert self.wf.data_input_ids == {"571", "572"}
        assert self.wf.source_ids == {"571", "572"}
        assert self.wf.sink_ids == {"573"}

    def test_dag(self):
        inv_dag: Dict[str, Set[str]] = {}
        for h, tails in self.wf.dag.items():
            for t in tails:
                inv_dag.setdefault(str(t), set()).add(h)
        assert self.wf.inv_dag == inv_dag
        heads = set(self.wf.dag)
        assert heads == set.union(*self.wf.inv_dag.values())
        tails = set(self.wf.inv_dag)
        assert tails == set.union(*self.wf.dag.values())
        ids = self.wf.sorted_step_ids()
        assert set(ids) == heads | tails
        for h, tails in self.wf.dag.items():
            for t in tails:
                assert ids.index(h) < ids.index(t)

    def test_steps(self):
        steps = SAMPLE_WF_DICT["steps"]
        assert isinstance(steps, dict)
        for sid, s in self.wf.steps.items():
            assert isinstance(s, wrappers.Step)
            assert s.id == sid
            assert sid in steps
            assert s.parent is self.wf
        assert self.wf.data_input_ids == {"571", "572"}
        assert self.wf.tool_ids == {"573"}

    def test_taint(self):
        assert not self.wf.is_modified
        self.wf.steps["571"].tool_id = "foo"
        assert self.wf.is_modified

    def test_input_map(self):
        history = wrappers.History({}, gi=self.gi)
        library = wrappers.Library({}, gi=self.gi)
        hda = wrappers.HistoryDatasetAssociation({"id": "hda_id"}, container=history, gi=self.gi)
        ldda = wrappers.LibraryDatasetDatasetAssociation({"id": "ldda_id"}, container=library, gi=self.gi)
        input_map = self.wf._convert_input_map({"0": hda, "1": ldda, "2": {"id": "hda2_id", "src": "hda"}})
        assert input_map == {
            "0": {"id": "hda_id", "src": "hda"},
            "1": {"id": "ldda_id", "src": "ldda"},
            "2": {"id": "hda2_id", "src": "hda"},
        }


@test_util.skip_unless_galaxy("release_19.09")
class TestInvocation(GalaxyObjectsTestBase):
    dataset: wrappers.HistoryDatasetAssociation
    history: wrappers.History
    inv: wrappers.Invocation
    workflow: wrappers.Workflow
    workflow_pause: wrappers.Workflow

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.inv = wrappers.Invocation(SAMPLE_INV_DICT, gi=cls.gi)
        with open(SAMPLE_FN) as f:
            cls.workflow = cls.gi.workflows.import_new(f.read())
        path_pause = test_util.get_abspath(os.path.join("data", "test_workflow_pause.ga"))
        with open(path_pause) as f:
            cls.workflow_pause = cls.gi.workflows.import_new(f.read())
        cls.history = cls.gi.histories.create(name="TestInvocation")
        cls.dataset = cls.history.paste_content("1\t2\t3")

    @classmethod
    def tearDownClass(cls):
        cls.history.delete(purge=True)

    def test_initialize(self):
        assert self.inv.workflow_id == "03501d7626bd192f"
        assert self.inv.history_id == "2f94e8ae9edff68a"
        assert self.inv.id == "df7a1f0c02a5b08e"
        assert self.inv.state == "ready"
        assert self.inv.update_time == "2015-10-31T22:00:26"
        assert self.inv.uuid == "c8aa2b1c-801a-11e5-a9e5-8ca98228593c"

    def test_initialize_steps(self):
        for step, step_dict in zip(self.inv.steps, SAMPLE_INV_DICT["steps"]):
            assert isinstance(step_dict, dict)
            assert isinstance(step, wrappers.InvocationStep)
            assert step.parent is self.inv
            assert step.id == step_dict["id"]
            assert step.job_id == step_dict["job_id"]
            assert step.order_index == step_dict["order_index"]
            assert step.state == step_dict["state"]
            assert step.update_time == step_dict["update_time"]
            assert step.workflow_step_id == step_dict["workflow_step_id"]
            assert step.workflow_step_label == step_dict["workflow_step_label"]
            assert step.workflow_step_uuid == step_dict["workflow_step_uuid"]

    def test_initialize_inputs(self):
        for i, input in enumerate(self.inv.inputs):
            assert input == {**SAMPLE_INV_DICT["inputs"][str(i)], "label": str(i)}

    def test_sorted_step_ids(self):
        assert self.inv.sorted_step_ids() == ["d413a19dec13d11e", "2f94e8ae9edff68a"]

    def test_step_states(self):
        assert self.inv.step_states() == {None, "new"}

    def test_number_of_steps(self):
        assert self.inv.number_of_steps() == 2

    def test_sorted_steps_by(self):
        assert len(self.inv.sorted_steps_by()) == 2
        steps = self.inv.sorted_steps_by(step_ids={"2f94e8ae9edff68a"})
        assert len(steps) == 1
        assert steps[0].id == "2f94e8ae9edff68a"
        assert self.inv.sorted_steps_by(step_ids={"unmatched_id"}) == []
        steps = self.inv.sorted_steps_by(states={"new"})
        assert len(steps) == 1
        assert steps[0].state == "new"
        assert self.inv.sorted_steps_by(states={"unmatched_state"}) == []
        steps = self.inv.sorted_steps_by(indices={0}, states={None, "new"})
        assert len(steps) == 1
        assert steps[0].order_index == 0
        assert self.inv.sorted_steps_by(indices={2}) == []

    def test_cancel(self):
        inv = self._obj_invoke_workflow()
        inv.cancel()
        assert inv.state in ["cancelled", "cancelling"]

    def test_wait(self):
        inv = self._obj_invoke_workflow()
        inv.wait()
        assert inv.state == "scheduled"

    def test_refresh(self):
        inv = self._obj_invoke_workflow()
        inv.state = "placeholder"
        # use wait_for_invocation() directly, because inv.wait() will update inv automatically
        self.gi.gi.invocations.wait_for_invocation(inv.id)
        inv.refresh()
        assert inv.state == "scheduled"

    def test_run_step_actions(self):
        inv = self.workflow_pause.invoke(
            inputs={"0": self.dataset},
            history=self.history,
        )
        for _ in range(20):
            with pytest.raises(bioblend.TimeoutException):
                inv.wait(maxwait=0.5, interval=0.5)
            inv.refresh()
            if len(inv.steps) >= 3:
                break
        assert inv.steps[2].action is None
        inv.run_step_actions([inv.steps[2]], [True])
        assert inv.steps[2].action is True

    def test_summary(self):
        inv = self._obj_invoke_workflow()
        inv.wait()
        summary = inv.summary()
        assert summary["populated_state"] == "ok"

    def test_step_jobs_summary(self):
        inv = self._obj_invoke_workflow()
        inv.wait()
        step_jobs_summary = inv.step_jobs_summary()
        assert len(step_jobs_summary) == 1
        assert step_jobs_summary[0]["populated_state"] == "ok"

    def test_report(self):
        inv = self._obj_invoke_workflow()
        report = inv.report()
        assert report["workflows"] == {self.workflow.id: {"name": "paste_columns"}}

    @test_util.skip_unless_galaxy("release_20.09")
    def test_biocompute_object(self):
        inv = self._obj_invoke_workflow()
        inv.wait()
        biocompute_object = inv.biocompute_object()
        assert len(biocompute_object["description_domain"]["pipeline_steps"]) == 1

    def _obj_invoke_workflow(self) -> wrappers.Invocation:
        return self.workflow.invoke(
            inputs={"Input 1": self.dataset, "Input 2": self.dataset},
            history=self.history,
            inputs_by="name",
        )


@test_util.skip_unless_galaxy("release_19.09")
class TestObjInvocationClient(GalaxyObjectsTestBase):
    history: wrappers.History
    inv: wrappers.Invocation
    workflow: wrappers.Workflow

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        with open(SAMPLE_FN) as f:
            cls.workflow = cls.gi.workflows.import_new(f.read())
        cls.history = cls.gi.histories.create(name="TestGalaxyObjInvocationClient")
        dataset = cls.history.paste_content("1\t2\t3")
        cls.inv = cls.workflow.invoke(
            inputs={"Input 1": dataset, "Input 2": dataset},
            history=cls.history,
            inputs_by="name",
        )
        cls.inv.wait()

    @classmethod
    def tearDownClass(cls):
        cls.history.delete(purge=True)

    def test_get(self):
        inv = self.gi.invocations.get(self.inv.id)
        assert inv.id == self.inv.id
        assert inv.workflow_id == self.workflow.id
        assert inv.history_id == self.history.id
        assert inv.state == "scheduled"
        assert inv.update_time == self.inv.update_time
        assert inv.uuid == self.inv.uuid

    def test_get_previews(self):
        previews = self.gi.invocations.get_previews()
        assert {type(preview) for preview in previews} == {wrappers.InvocationPreview}
        inv_preview = next(p for p in previews if p.id == self.inv.id)
        assert inv_preview.id == self.inv.id
        assert inv_preview.workflow_id == self.workflow.id
        assert inv_preview.history_id == self.history.id
        assert inv_preview.state == "scheduled"
        assert inv_preview.update_time == self.inv.update_time
        assert inv_preview.uuid == self.inv.uuid

    def test_list(self):
        invs = self.gi.invocations.list()
        inv = next(i for i in invs if i.id == self.inv.id)
        assert inv.id == self.inv.id
        assert inv.workflow_id == self.workflow.id
        assert inv.history_id == self.history.id
        assert inv.state == "scheduled"
        assert inv.update_time == self.inv.update_time
        assert inv.uuid == self.inv.uuid
        assert len(self.inv.steps) > 0
        history = self.gi.histories.create(name="TestGalaxyObjInvocationClientList")
        assert self.gi.invocations.list(history=history) == []
        history.delete(purge=True)


class TestGalaxyInstance(GalaxyObjectsTestBase):
    def test_library(self):
        name = f"test_{uuid.uuid4().hex}"
        description, synopsis = "D", "S"
        lib = self.gi.libraries.create(name, description=description, synopsis=synopsis)
        assert lib.name == name
        assert lib.description == description
        assert lib.synopsis == synopsis
        assert len(lib.content_infos) == 1  # root folder
        assert len(lib.folder_ids) == 1
        assert len(lib.dataset_ids) == 0
        assert lib.id in [_.id for _ in self.gi.libraries.list()]
        lib.delete()
        assert not lib.is_mapped

    def test_workflow_from_str(self):
        with open(SAMPLE_FN) as f:
            wf = self.gi.workflows.import_new(f.read())
        self._check_and_del_workflow(wf)

    def test_workflow_collections_from_str(self):
        with open(SAMPLE_WF_COLL_FN) as f:
            wf = self.gi.workflows.import_new(f.read())
        self._check_and_del_workflow(wf)

    def test_workflow_parameter_input(self):
        with open(SAMPLE_WF_PARAMETER_INPUT_FN) as f:
            self.gi.workflows.import_new(f.read())

    def test_workflow_from_dict(self):
        with open(SAMPLE_FN) as f:
            wf = self.gi.workflows.import_new(json.load(f))
        self._check_and_del_workflow(wf)

    def test_workflow_publish_from_dict(self):
        with open(SAMPLE_FN) as f:
            wf = self.gi.workflows.import_new(json.load(f), publish=True)
        self._check_and_del_workflow(wf, check_is_public=True)

    def test_workflow_missing_tools(self):
        with open(SAMPLE_FN) as f:
            wf_dump = json.load(f)
        wf_info = self.gi.gi.workflows.import_workflow_dict(wf_dump)
        wf_dict = self.gi.gi.workflows.show_workflow(wf_info["id"])
        for id_, step in wf_dict["steps"].items():
            if step["type"] == "tool":
                for k in "tool_inputs", "tool_version":
                    wf_dict["steps"][id_][k] = None
        wf = wrappers.Workflow(wf_dict, gi=self.gi)
        assert not wf.is_runnable
        with pytest.raises(RuntimeError):
            wf.invoke()
        wf.delete()

    def test_workflow_export(self):
        with open(SAMPLE_FN) as f:
            wf1 = self.gi.workflows.import_new(f.read())
        wf2 = self.gi.workflows.import_new(wf1.export())
        assert wf1.id != wf2.id
        for wf in wf1, wf2:
            self._check_and_del_workflow(wf)

    def _check_and_del_workflow(self, wf: wrappers.Workflow, check_is_public: bool = False) -> None:
        # Galaxy appends additional text to imported workflow names
        assert wf.name.startswith("paste_columns")
        assert len(wf.steps) == 3
        for step_id, step in wf.steps.items():
            assert isinstance(step, wrappers.Step)
            assert step_id == step.id
            assert isinstance(step.tool_inputs, dict)
            if step.type == "tool":
                assert step.tool_id is not None
                assert step.tool_version is not None
                assert isinstance(step.input_steps, dict)
            elif step.type in ("data_collection_input", "data_input"):
                assert step.tool_id is None
                assert step.tool_version is None
                assert step.input_steps == {}
        wf_ids = {_.id for _ in self.gi.workflows.list()}
        assert wf.id in wf_ids
        if check_is_public:
            assert wf.published
        wf.delete()

    # not very accurate:
    #   * we can't publish a wf from the API
    #   * we can't directly get another user's wf
    def test_workflow_from_shared(self):
        all_prevs = {_.id: _ for _ in self.gi.workflows.get_previews(published=True)}
        pub_only_ids = set(all_prevs).difference(_.id for _ in self.gi.workflows.get_previews())
        if pub_only_ids:
            wf_id = pub_only_ids.pop()
            imported = self.gi.workflows.import_shared(wf_id)
            assert isinstance(imported, wrappers.Workflow)
            imported.delete()
        else:
            self.skipTest("no published workflows, manually publish a workflow to run this test")

    def test_get_libraries(self):
        self._test_multi_get("libraries")

    def test_get_histories(self):
        self._test_multi_get("histories")

    def test_get_workflows(self):
        self._test_multi_get("workflows")

    def _normalized_functions(
        self, obj_type: Literal["histories", "libraries", "workflows"]
    ) -> Tuple[Callable, Dict[str, Any]]:
        if obj_type == "libraries":
            create: Callable = self.gi.libraries.create
            del_kwargs = {}
        elif obj_type == "histories":
            create = self.gi.histories.create
            del_kwargs = {"purge": True}
        elif obj_type == "workflows":

            def create(name):
                with open(SAMPLE_FN) as f:
                    d = json.load(f)
                d["name"] = name
                return self.gi.workflows.import_new(d)

            del_kwargs = {}
        return create, del_kwargs

    def _test_multi_get(self, obj_type: Literal["histories", "libraries", "workflows"]) -> None:
        obj_gi_client = getattr(self.gi, obj_type)
        create, del_kwargs = self._normalized_functions(obj_type)

        def ids(seq: Iterable[wrappers.Wrapper]) -> Set[str]:
            return {_.id for _ in seq}

        names = [f"test_{uuid.uuid4().hex}" for _ in range(2)]
        objs = []
        try:
            objs = [create(_) for _ in names]
            assert ids(objs) <= ids(obj_gi_client.list())
            if obj_type != "workflows":
                filtered = obj_gi_client.list(name=names[0])
                assert len(filtered) == 1
                assert filtered[0].id == objs[0].id
                del_id = objs[-1].id
                objs.pop().delete(**del_kwargs)
                assert del_id in ids(obj_gi_client.get_previews(deleted=True))
            else:
                # Galaxy appends info strings to imported workflow names
                prev = obj_gi_client.get_previews()[0]
                filtered = obj_gi_client.list(name=prev.name)
                assert len(filtered) == 1
                assert filtered[0].id == prev.id
        finally:
            for o in objs:
                o.delete(**del_kwargs)

    def test_delete_libraries_by_name(self):
        self._test_delete_by_name("libraries")
        self._test_delete_by_ambiguous_name("libraries")

    def test_delete_histories_by_name(self):
        self._test_delete_by_name("histories")
        self._test_delete_by_ambiguous_name("histories")

    def test_delete_workflows_by_name(self):
        self._test_delete_by_name("workflows")
        self._test_delete_by_ambiguous_name("workflows")

    def _test_delete_by_name(self, obj_type: Literal["histories", "libraries", "workflows"]) -> None:
        obj_gi_client = getattr(self.gi, obj_type)
        create, del_kwargs = self._normalized_functions(obj_type)
        name = f"test_{uuid.uuid4().hex}"
        create(name)
        prevs = [_ for _ in obj_gi_client.get_previews(name=name) if not _.deleted]
        assert len(prevs) == 1
        del_kwargs["name"] = name
        obj_gi_client.delete(**del_kwargs)
        prevs = [_ for _ in obj_gi_client.get_previews(name=name) if not _.deleted]
        assert len(prevs) == 0

    def _test_delete_by_ambiguous_name(self, obj_type: Literal["histories", "libraries", "workflows"]) -> None:
        obj_gi_client = getattr(self.gi, obj_type)
        create, del_kwargs = self._normalized_functions(obj_type)
        name = f"test_{uuid.uuid4().hex}"
        objs = [create(name) for _ in range(2)]
        prevs = [_ for _ in obj_gi_client.get_previews(name=name) if not _.deleted]
        assert len(prevs) == len(objs)
        del_kwargs["name"] = name
        with pytest.raises(ValueError):
            obj_gi_client.delete(**del_kwargs)
        # Cleanup
        del del_kwargs["name"]
        for prev in prevs:
            del_kwargs["id_"] = prev.id
            obj_gi_client.delete(**del_kwargs)


class TestLibrary(GalaxyObjectsTestBase):
    # just something that can be expected to be always up
    DS_URL = "https://tools.ietf.org/rfc/rfc1866.txt"

    def setUp(self):
        super().setUp()
        self.lib = self.gi.libraries.create(f"test_{uuid.uuid4().hex}")

    def tearDown(self):
        self.lib.delete()

    def test_root_folder(self):
        r = self.lib.root_folder
        assert r.parent is None

    def test_folder(self):
        name = f"test_{uuid.uuid4().hex}"
        desc = "D"
        folder = self.lib.create_folder(name, description=desc)
        assert folder.name == name
        assert folder.description == desc
        assert folder.container is self.lib
        assert folder.parent is not None
        assert folder.parent.id == self.lib.root_folder.id
        assert len(self.lib.content_infos) == 2
        assert len(self.lib.folder_ids) == 2
        assert folder.id in self.lib.folder_ids
        retrieved = self.lib.get_folder(folder.id)
        assert folder.id == retrieved.id

    def _check_datasets(self, dss: Collection[wrappers.LibraryDataset]) -> None:
        assert len(dss) == len(self.lib.dataset_ids)
        assert {_.id for _ in dss} == set(self.lib.dataset_ids)
        for ds in dss:
            assert ds.container is self.lib

    def test_dataset(self):
        folder = self.lib.create_folder(f"test_{uuid.uuid4().hex}")
        ds = self.lib.upload_data(FOO_DATA, folder=folder)
        assert len(self.lib.content_infos) == 3
        assert len(self.lib.folder_ids) == 2
        self._check_datasets([ds])

    def test_dataset_from_url(self):
        if is_reachable(self.DS_URL):
            ds = self.lib.upload_from_url(self.DS_URL)
            self._check_datasets([ds])
        else:
            self.skipTest(f"{self.DS_URL} not reachable")

    def test_dataset_from_local(self):
        with tempfile.NamedTemporaryFile(mode="w", prefix="bioblend_test_") as f:
            f.write(FOO_DATA)
            f.flush()
            ds = self.lib.upload_from_local(f.name)
        self._check_datasets([ds])

    def test_datasets_from_fs(self):
        bnames = [f"f{i}.txt" for i in range(2)]
        dss, fnames = upload_from_fs(self.lib, bnames)
        self._check_datasets(dss)
        dss, fnames = upload_from_fs(self.lib, bnames, link_data_only="link_to_files")
        for ds, fn in zip(dss, fnames):
            assert ds.file_name == fn

    def test_copy_from_dataset(self):
        hist = self.gi.histories.create(f"test_{uuid.uuid4().hex}")
        try:
            hda = hist.paste_content(FOO_DATA)
            ds = self.lib.copy_from_dataset(hda)
        finally:
            hist.delete(purge=True)
        self._check_datasets([ds])

    def test_get_dataset(self):
        ds = self.lib.upload_data(FOO_DATA)
        retrieved = self.lib.get_dataset(ds.id)
        assert ds.id == retrieved.id

    def test_get_datasets(self):
        bnames = [f"f{i}.txt" for i in range(2)]
        dss, _ = upload_from_fs(self.lib, bnames)
        retrieved = self.lib.get_datasets()
        assert len(dss) == len(retrieved)
        assert {_.id for _ in dss} == {_.id for _ in retrieved}
        name = f"/{bnames[0]}"
        selected = self.lib.get_datasets(name=name)
        assert len(selected) == 1
        assert selected[0].name == bnames[0]


class TestLDContents(GalaxyObjectsTestBase):
    def setUp(self):
        super().setUp()
        self.lib = self.gi.libraries.create(f"test_{uuid.uuid4().hex}")
        self.ds = self.lib.upload_data(FOO_DATA)
        self.ds.wait()

    def tearDown(self):
        self.lib.delete()

    def test_dataset_get_stream(self):
        for idx, c in enumerate(self.ds.get_stream(chunk_size=1)):
            assert FOO_DATA[idx].encode() == c

    def test_dataset_peek(self):
        fetched_data = self.ds.peek(chunk_size=4)
        assert FOO_DATA[0:4].encode() == fetched_data

    def test_dataset_download(self):
        with tempfile.TemporaryFile() as f:
            self.ds.download(f)
            f.seek(0)
            assert FOO_DATA.encode() == f.read()

    def test_dataset_get_contents(self):
        assert FOO_DATA.encode() == self.ds.get_contents()

    def test_dataset_delete(self):
        self.ds.delete()
        # Cannot test this yet because the 'deleted' attribute is not exported
        # by the API at the moment
        # assert self.ds.deleted

    def test_dataset_update(self):
        new_name = f"test_{uuid.uuid4().hex}"
        new_misc_info = f"Annotation for {new_name}"
        new_genome_build = "hg19"
        updated_ldda = self.ds.update(name=new_name, misc_info=new_misc_info, genome_build=new_genome_build)
        assert self.ds.id == updated_ldda.id
        assert self.ds.name == new_name
        assert self.ds.misc_info == new_misc_info
        assert self.ds.genome_build == new_genome_build


class TestHistory(GalaxyObjectsTestBase):
    def setUp(self):
        super().setUp()
        self.hist = self.gi.histories.create(f"test_{uuid.uuid4().hex}")

    def tearDown(self):
        self.hist.delete(purge=True)

    def test_create_delete(self):
        name = f"test_{uuid.uuid4().hex}"
        hist = self.gi.histories.create(name)
        assert hist.name == name
        hist_id = hist.id
        assert hist_id in [_.id for _ in self.gi.histories.list()]
        hist.delete(purge=True)
        assert not hist.is_mapped
        h = self.gi.histories.get(hist_id)
        assert h.deleted

    def _check_dataset(self, hda: wrappers.HistoryDatasetAssociation) -> None:
        assert isinstance(hda, wrappers.HistoryDatasetAssociation)
        assert hda.container is self.hist
        assert len(self.hist.dataset_ids) == 1
        assert self.hist.dataset_ids[0] == hda.id

    def test_import_dataset(self):
        lib = self.gi.libraries.create(f"test_{uuid.uuid4().hex}")
        lds = lib.upload_data(FOO_DATA)
        assert len(self.hist.dataset_ids) == 0
        hda = self.hist.import_dataset(lds)
        lib.delete()
        self._check_dataset(hda)

    def test_upload_file(self):
        with tempfile.NamedTemporaryFile(mode="w", prefix="bioblend_test_") as f:
            f.write(FOO_DATA)
            f.flush()
            hda = self.hist.upload_file(f.name)
        self._check_dataset(hda)

    def test_paste_content(self):
        hda = self.hist.paste_content(FOO_DATA)
        self._check_dataset(hda)

    def test_get_dataset(self):
        hda = self.hist.paste_content(FOO_DATA)
        retrieved = self.hist.get_dataset(hda.id)
        assert hda.id == retrieved.id

    def test_get_datasets(self):
        bnames = [f"f{i}.txt" for i in range(2)]
        lib = self.gi.libraries.create(f"test_{uuid.uuid4().hex}")
        lds = upload_from_fs(lib, bnames)[0]
        hdas = [self.hist.import_dataset(_) for _ in lds]
        lib.delete()
        retrieved = self.hist.get_datasets()
        assert len(hdas) == len(retrieved)
        assert {_.id for _ in hdas} == {_.id for _ in retrieved}
        selected = self.hist.get_datasets(name=bnames[0])
        assert len(selected) == 1
        assert selected[0].name == bnames[0]

    def test_export_and_download(self):
        jeha_id = self.hist.export(wait=True, maxwait=60)
        assert jeha_id
        tempdir = tempfile.mkdtemp(prefix="bioblend_test_")
        temp_fn = os.path.join(tempdir, "export.tar.gz")
        try:
            with open(temp_fn, "wb") as fo:
                self.hist.download(jeha_id, fo)
            assert tarfile.is_tarfile(temp_fn)
        finally:
            shutil.rmtree(tempdir)

    def test_update(self):
        new_name = f"test_{uuid.uuid4().hex}"
        new_annotation = f"Annotation for {new_name}"
        new_tags = ["tag1", "tag2"]
        updated_hist = self.hist.update(name=new_name, annotation=new_annotation, tags=new_tags)
        assert self.hist.id == updated_hist.id
        assert self.hist.name == new_name
        assert self.hist.annotation == new_annotation
        assert self.hist.tags == new_tags
        updated_hist = self.hist.update(published=True)
        assert self.hist.id == updated_hist.id
        assert self.hist.published

    def test_create_dataset_collection(self):
        self._create_collection_description()
        hdca = self.hist.create_dataset_collection(self.collection_description, copy_elements=False)
        assert isinstance(hdca, wrappers.HistoryDatasetCollectionAssociation)
        assert hdca.collection_type == "list"
        assert hdca.container is self.hist
        assert len(hdca.elements) == 2
        assert self.dataset1.id == hdca.elements[0]["object"]["id"]
        assert self.dataset2.id == hdca.elements[1]["object"]["id"]

    def test_delete_dataset_collection(self):
        self._create_collection_description()
        hdca = self.hist.create_dataset_collection(self.collection_description, copy_elements=False)
        hdca.delete()
        assert hdca.deleted

    def _create_collection_description(self) -> None:
        self.dataset1 = self.hist.paste_content(FOO_DATA)
        self.dataset2 = self.hist.paste_content(FOO_DATA_2)
        self.collection_description = dataset_collections.CollectionDescription(
            name="MyDatasetList",
            elements=[
                dataset_collections.HistoryDatasetElement(name="sample1", id=self.dataset1.id),
                dataset_collections.HistoryDatasetElement(name="sample2", id=self.dataset2.id),
            ],
        )


class TestHDAContents(GalaxyObjectsTestBase):
    def setUp(self):
        super().setUp()
        self.hist = self.gi.histories.create(f"test_{uuid.uuid4().hex}")
        self.ds = self.hist.paste_content(FOO_DATA)

    def tearDown(self):
        self.hist.delete(purge=True)

    def test_dataset_get_stream(self):
        for idx, c in enumerate(self.ds.get_stream(chunk_size=1)):
            assert FOO_DATA[idx].encode() == c

    def test_dataset_peek(self):
        fetched_data = self.ds.peek(chunk_size=4)
        assert FOO_DATA[0:4].encode() == fetched_data

    def test_dataset_download(self):
        with tempfile.TemporaryFile() as f:
            self.ds.download(f)
            f.seek(0)
            assert FOO_DATA.encode() == f.read()

    def test_dataset_get_contents(self):
        assert FOO_DATA.encode() == self.ds.get_contents()

    def test_dataset_update(self):
        new_name = f"test_{uuid.uuid4().hex}"
        new_annotation = f"Annotation for {new_name}"
        new_genome_build = "hg19"
        updated_hda = self.ds.update(name=new_name, annotation=new_annotation, genome_build=new_genome_build)
        assert self.ds.id == updated_hda.id
        assert self.ds.name == new_name
        assert self.ds.annotation == new_annotation
        assert self.ds.genome_build == new_genome_build

    def test_dataset_delete(self):
        self.ds.delete()
        assert self.ds.deleted
        assert not self.ds.purged

    def test_dataset_purge(self):
        self.ds.delete(purge=True, wait=True)
        assert self.ds.deleted
        assert self.ds.purged


@test_util.skip_unless_galaxy("release_19.09")
class TestRunWorkflow(GalaxyObjectsTestBase):
    def setUp(self):
        super().setUp()
        self.lib = self.gi.libraries.create(f"test_{uuid.uuid4().hex}")
        with open(SAMPLE_FN) as f:
            self.wf = self.gi.workflows.import_new(f.read())
        self.contents = ["one\ntwo\n", "1\n2\n"]
        self.inputs = [self.lib.upload_data(_) for _ in self.contents]

    def tearDown(self):
        self.wf.delete()
        self.lib.delete()

    def _test(self, existing_hist: bool = False, pass_params: bool = False) -> None:
        hist_name = f"test_{uuid.uuid4().hex}"
        if existing_hist:
            hist: Union[str, wrappers.History] = self.gi.histories.create(hist_name)
        else:
            hist = hist_name
        if pass_params:
            params = {"Paste1": {"delimiter": "U"}}
            sep = "_"  # 'U' maps to '_' in the paste tool
        else:
            params = None
            sep = "\t"  # default
        input_map = {"Input 1": self.inputs[0], "Input 2": self.inputs[1]}
        sys.stderr.write(os.linesep)
        inv = self.wf.invoke(inputs=input_map, params=params, history=hist, inputs_by="name")
        out_hist = self.gi.histories.get(inv.history_id)
        inv.wait()
        last_step = inv.sorted_steps_by()[-1]
        out_ds = last_step.get_outputs()["out_file1"]
        assert out_ds.container.id == out_hist.id
        res = out_ds.get_contents()
        exp_rows = zip(*(_.splitlines() for _ in self.contents))
        exp_res = ("\n".join(sep.join(t) for t in exp_rows) + "\n").encode()
        assert res == exp_res
        if isinstance(hist, wrappers.History):  # i.e. existing_hist == True
            assert out_hist.id == hist.id
        out_hist.delete(purge=True)

    def test_existing_history(self) -> None:
        self._test(existing_hist=True)

    def test_new_history(self) -> None:
        self._test(existing_hist=False)

    def test_params(self) -> None:
        self._test(pass_params=True)


@test_util.skip_unless_galaxy("release_19.09")
class TestRunDatasetCollectionWorkflow(GalaxyObjectsTestBase):
    def setUp(self):
        super().setUp()
        with open(SAMPLE_WF_COLL_FN) as f:
            self.wf = self.gi.workflows.import_new(f.read())
        self.hist = self.gi.histories.create(f"test_{uuid.uuid4().hex}")

    def tearDown(self):
        self.wf.delete()
        self.hist.delete(purge=True)

    def test_run_workflow_with_dataset_collection(self):
        dataset1 = self.hist.paste_content(FOO_DATA)
        dataset2 = self.hist.paste_content(FOO_DATA_2)
        collection_description = dataset_collections.CollectionDescription(
            name="MyDatasetList",
            elements=[
                dataset_collections.HistoryDatasetElement(name="sample1", id=dataset1.id),
                dataset_collections.HistoryDatasetElement(name="sample2", id=dataset2.id),
            ],
        )
        dataset_collection = self.hist.create_dataset_collection(collection_description, copy_elements=False)
        assert len(self.hist.content_infos) == 3
        input_map = {"0": dataset_collection, "1": dataset1}
        inv = self.wf.invoke(input_map, history=self.hist)
        inv.wait()
        self.hist.refresh()
        assert len(self.hist.content_infos) == 6
        last_step = inv.sorted_steps_by()[-1]
        out_hdca = last_step.get_output_collections()["out_file1"]
        assert out_hdca.collection_type == "list"
        assert len(out_hdca.elements) == 2
        assert out_hdca.container.id == self.hist.id


class TestJob(GalaxyObjectsTestBase):
    def test_get(self):
        job_prevs = self.gi.jobs.get_previews()
        if len(job_prevs) > 0:
            job_prev = job_prevs[0]
            assert isinstance(job_prev, wrappers.JobPreview)
            job = self.gi.jobs.get(job_prev.id)
            assert isinstance(job, wrappers.Job)
            assert job.id == job_prev.id
        for job in self.gi.jobs.list():
            assert isinstance(job, wrappers.Job)



"""
Contains possible interactions with the Galaxy workflow invocations
"""

import logging
from typing import (
    Any,
    Dict,
    List,
    Optional,
    TYPE_CHECKING,
)

from bioblend import (
    CHUNK_SIZE,
    ConnectionError,
    NotReady,
    wait_on,
)
from bioblend.galaxy.client import Client
from bioblend.galaxy.workflows import InputsBy

if TYPE_CHECKING:
    from bioblend.galaxy import GalaxyInstance

log = logging.getLogger(__name__)

INVOCATION_TERMINAL_STATES = {"cancelled", "failed", "scheduled"}
# Invocation non-terminal states are: "cancelling", "new", "ready"


class InvocationClient(Client):
    gi: "GalaxyInstance"
    module = "invocations"

    def __init__(self, galaxy_instance: "GalaxyInstance") -> None:
        super().__init__(galaxy_instance)

    def get_invocations(
        self,
        workflow_id: Optional[str] = None,
        history_id: Optional[str] = None,
        user_id: Optional[str] = None,
        include_terminal: bool = True,
        limit: Optional[int] = None,
        view: str = "collection",
        step_details: bool = False,
    ) -> List[Dict[str, Any]]:
        """
        Get all workflow invocations, or select a subset by specifying optional
        arguments for filtering (e.g. a workflow ID).

        :type workflow_id: str
        :param workflow_id: Encoded workflow ID to filter on

        :type history_id: str
        :param history_id: Encoded history ID to filter on

        :type user_id: str
        :param user_id: Encoded user ID to filter on. This must be
                        your own user ID if your are not an admin user.

        :type include_terminal: bool
        :param include_terminal: Whether to include terminal states.

        :type limit: int
        :param limit: Maximum number of invocations to return - if specified,
                      the most recent invocations will be returned.

        :type view: str
        :param view: Level of detail to return per invocation, either
                     'element' or 'collection'.

        :type step_details: bool
        :param step_details: If 'view' is 'element', also include details
                             on individual steps.

        :rtype: list
        :return: A list of workflow invocations.
          For example::

            [{'history_id': '2f94e8ae9edff68a',
              'id': 'df7a1f0c02a5b08e',
              'model_class': 'WorkflowInvocation',
              'state': 'new',
              'update_time': '2015-10-31T22:00:22',
              'uuid': 'c8aa2b1c-801a-11e5-a9e5-8ca98228593c',
              'workflow_id': '03501d7626bd192f'}]
        """
        params = {"include_terminal": include_terminal, "view": view, "step_details": step_details}
        if workflow_id:
            params["workflow_id"] = workflow_id
        if history_id:
            params["history_id"] = history_id
        if user_id:
            params["user_id"] = user_id
        if limit is not None:
            params["limit"] = limit
        return self._get(params=params)

    def show_invocation(self, invocation_id: str) -> Dict[str, Any]:
        """
        Get a workflow invocation dictionary representing the scheduling of a
        workflow. This dictionary may be sparse at first (missing inputs and
        invocation steps) and will become more populated as the workflow is
        actually scheduled.

        :type invocation_id: str
        :param invocation_id: Encoded workflow invocation ID

        :rtype: dict
        :return: The workflow invocation.
          For example::

            {'history_id': '2f94e8ae9edff68a',
             'id': 'df7a1f0c02a5b08e',
             'inputs': {'0': {'id': 'a7db2fac67043c7e',
               'src': 'hda',
               'uuid': '7932ffe0-2340-4952-8857-dbaa50f1f46a'}},
             'model_class': 'WorkflowInvocation',
             'state': 'ready',
             'steps': [{'action': None,
               'id': 'd413a19dec13d11e',
               'job_id': None,
               'model_class': 'WorkflowInvocationStep',
               'order_index': 0,
               'state': None,
               'update_time': '2015-10-31T22:00:26',
               'workflow_step_id': 'cbbbf59e8f08c98c',
               'workflow_step_label': None,
               'workflow_step_uuid': 'b81250fd-3278-4e6a-b269-56a1f01ef485'},
              {'action': None,
               'id': '2f94e8ae9edff68a',
               'job_id': 'e89067bb68bee7a0',
               'model_class': 'WorkflowInvocationStep',
               'order_index': 1,
               'state': 'new',
               'update_time': '2015-10-31T22:00:26',
               'workflow_step_id': '964b37715ec9bd22',
               'workflow_step_label': None,
               'workflow_step_uuid': 'e62440b8-e911-408b-b124-e05435d3125e'}],
             'update_time': '2015-10-31T22:00:26',
             'uuid': 'c8aa2b1c-801a-11e5-a9e5-8ca98228593c',
             'workflow_id': '03501d7626bd192f'}
        """
        url = self._make_url(invocation_id)
        return self._get(url=url)

    def rerun_invocation(
        self,
        invocation_id: str,
        inputs_update: Optional[dict] = None,
        params_update: Optional[dict] = None,
        history_id: Optional[str] = None,
        history_name: Optional[str] = None,
        import_inputs_to_history: bool = False,
        replacement_params: Optional[dict] = None,
        allow_tool_state_corrections: bool = False,
        inputs_by: Optional[InputsBy] = None,
        parameters_normalized: bool = False,
    ) -> Dict[str, Any]:
        """
        Rerun a workflow invocation. For more extensive documentation of all
        parameters, see the ``gi.workflows.invoke_workflow()`` method.

        :type invocation_id: str
        :param invocation_id: Encoded workflow invocation ID to be rerun

        :type inputs_update: dict
        :param inputs_update: If different datasets should be used to the original
          invocation, this should contain a mapping of workflow inputs to the new
          datasets and dataset collections.

        :type params_update: dict
        :param params_update: If different non-dataset tool parameters should be
          used to the original invocation, this should contain a mapping of the
          new parameter values.

        :type history_id: str
        :param history_id: The encoded history ID where to store the workflow
          outputs. Alternatively, ``history_name`` may be specified to create a
          new history.

        :type history_name: str
        :param history_name: Create a new history with the given name to store
          the workflow outputs. If both ``history_id`` and ``history_name`` are
          provided, ``history_name`` is ignored. If neither is specified, a new
          'Unnamed history' is created.

        :type import_inputs_to_history: bool
        :param import_inputs_to_history: If ``True``, used workflow inputs will
          be imported into the history. If ``False``, only workflow outputs will
          be visible in the given history.

        :type allow_tool_state_corrections: bool
        :param allow_tool_state_corrections: If True, allow Galaxy to fill in
          missing tool state when running workflows. This may be useful for
          workflows using tools that have changed over time or for workflows
          built outside of Galaxy with only a subset of inputs defined.

        :type replacement_params: dict
        :param replacement_params: pattern-based replacements for post-job
          actions

        :type inputs_by: str
        :param inputs_by: Determines how inputs are referenced. Can be
          "step_index|step_uuid" (default), "step_index", "step_id", "step_uuid", or "name".

        :type parameters_normalized: bool
        :param parameters_normalized: Whether Galaxy should normalize the input
          parameters to ensure everything is referenced by a numeric step ID.
          Default is ``False``, but when setting parameters for a subworkflow,
          ``True`` is required.

        :rtype: dict
        :return: A dict describing the new workflow invocation.

        .. note::
          This method works only on Galaxy 21.01 or later.
        """
        invocation_details = self.show_invocation(invocation_id)
        workflow_id = invocation_details["workflow_id"]
        inputs = invocation_details["inputs"]
        wf_params = invocation_details["input_step_parameters"]
        if inputs_update:
            for inp, input_value in inputs_update.items():
                inputs[inp] = input_value
        if params_update:
            for param, param_value in params_update.items():
                wf_params[param] = param_value
        payload = {"inputs": inputs, "params": wf_params}

        if replacement_params:
            payload["replacement_params"] = replacement_params
        if history_id:
            payload["history"] = f"hist_id={history_id}"
        elif history_name:
            payload["history"] = history_name
        if not import_inputs_to_history:
            payload["no_add_to_history"] = True
        if allow_tool_state_corrections:
            payload["allow_tool_state_corrections"] = allow_tool_state_corrections
        if inputs_by is not None:
            payload["inputs_by"] = inputs_by
        if parameters_normalized:
            payload["parameters_normalized"] = parameters_normalized
        api_params = {"instance": True}
        url = "/".join((self.gi.url, "workflows", workflow_id, "invocations"))
        return self.gi.make_post_request(url=url, payload=payload, params=api_params)

    def cancel_invocation(self, invocation_id: str) -> Dict[str, Any]:
        """
        Cancel the scheduling of a workflow.

        :type invocation_id: str
        :param invocation_id: Encoded workflow invocation ID

        :rtype: dict
        :return: The workflow invocation being cancelled
        """
        url = self._make_url(invocation_id)
        return self._delete(url=url)

    def show_invocation_step(self, invocation_id: str, step_id: str) -> Dict[str, Any]:
        """
        See the details of a particular workflow invocation step.

        :type invocation_id: str
        :param invocation_id: Encoded workflow invocation ID

        :type step_id: str
        :param step_id: Encoded workflow invocation step ID

        :rtype: dict
        :return: The workflow invocation step.
          For example::

            {'action': None,
             'id': '63cd3858d057a6d1',
             'job_id': None,
             'model_class': 'WorkflowInvocationStep',
             'order_index': 2,
             'state': None,
             'update_time': '2015-10-31T22:11:14',
             'workflow_step_id': '52e496b945151ee8',
             'workflow_step_label': None,
             'workflow_step_uuid': '4060554c-1dd5-4287-9040-8b4f281cf9dc'}
        """
        url = self._invocation_step_url(invocation_id, step_id)
        return self._get(url=url)

    def run_invocation_step_action(self, invocation_id: str, step_id: str, action: Any) -> Dict[str, Any]:
        """Execute an action for an active workflow invocation step. The
        nature of this action and what is expected will vary based on the
        the type of workflow step (the only currently valid action is True/False
        for pause steps).

        :type invocation_id: str
        :param invocation_id: Encoded workflow invocation ID

        :type step_id: str
        :param step_id: Encoded workflow invocation step ID

        :type action: object
        :param action: Action to use when updating state, semantics depends on
           step type.

        :rtype: dict
        :return: Representation of the workflow invocation step
        """
        url = self._invocation_step_url(invocation_id, step_id)
        payload = {"action": action}
        return self._put(payload=payload, url=url)

    def get_invocation_summary(self, invocation_id: str) -> Dict[str, Any]:
        """
        Get a summary of an invocation, stating the number of jobs which
        succeed, which are paused and which have errored.

        :type invocation_id: str
        :param invocation_id: Encoded workflow invocation ID

        :rtype: dict
        :return: The invocation summary.
          For example::

            {'states': {'paused': 4, 'error': 2, 'ok': 2},
             'model': 'WorkflowInvocation',
             'id': 'a799d38679e985db',
             'populated_state': 'ok'}
        """
        url = self._make_url(invocation_id) + "/jobs_summary"
        return self._get(url=url)

    def get_invocation_step_jobs_summary(self, invocation_id: str) -> List[Dict[str, Any]]:
        """
        Get a detailed summary of an invocation, listing all jobs with
        their job IDs and current states.

        :type invocation_id: str
        :param invocation_id: Encoded workflow invocation ID

        :rtype: list of dicts
        :return: The invocation step jobs summary.
          For example::

            [{'id': 'e85a3be143d5905b',
              'model': 'Job',
              'populated_state': 'ok',
              'states': {'ok': 1}},
             {'id': 'c9468fdb6dc5c5f1',
              'model': 'Job',
              'populated_state': 'ok',
              'states': {'running': 1}},
             {'id': '2a56795cad3c7db3',
              'model': 'Job',
              'populated_state': 'ok',
              'states': {'new': 1}}]
        """
        url = self._make_url(invocation_id) + "/step_jobs_summary"
        return self._get(url=url)

    def get_invocation_report(self, invocation_id: str) -> Dict[str, Any]:
        """
        Get a Markdown report for an invocation.

        :type invocation_id: str
        :param invocation_id: Encoded workflow invocation ID

        :rtype: dict
        :return: The invocation report.
          For example::

            {'markdown': '\\n# Workflow Execution Summary of Example workflow\\n\\n
             ## Workflow Inputs\\n\\n\\n## Workflow Outputs\\n\\n\\n
             ## Workflow\\n```galaxy\\n
             workflow_display(workflow_id=f2db41e1fa331b3e)\\n```\\n',
             'render_format': 'markdown',
             'workflows': {'f2db41e1fa331b3e': {'name': 'Example workflow'}}}
        """
        url = self._make_url(invocation_id) + "/report"
        return self._get(url=url)

    def get_invocation_report_pdf(self, invocation_id: str, file_path: str, chunk_size: int = CHUNK_SIZE) -> None:
        """
        Get a PDF report for an invocation.

        :type invocation_id: str
        :param invocation_id: Encoded workflow invocation ID

        :type file_path: str
        :param file_path: Path to save the report
        """
        url = self._make_url(invocation_id) + "/report.pdf"
        r = self.gi.make_get_request(url, stream=True)
        if r.status_code != 200:
            raise Exception(
                "Failed to get the PDF report, the necessary dependencies may not be installed on the Galaxy server."
            )
        with open(file_path, "wb") as outf:
            for chunk in r.iter_content(chunk_size):
                outf.write(chunk)

    # TODO: Move to a new ``bioblend.galaxy.short_term_storage`` module
    def _wait_for_short_term_storage(
        self, storage_request_id: str, maxwait: float = 60, interval: float = 3
    ) -> Dict[str, Any]:
        """
        Wait until a short term storage request is ready

        :type storage_request_id: str
        :param storage_request_id: Storage request ID to wait for.

        :type maxwait: float
        :param maxwait: Total time (in seconds) to wait for the storage request
          to become ready. After this time, a ``TimeoutException`` will be
          raised.

        :type interval: float
        :param interval: Time (in seconds) to wait between 2 consecutive checks.

        :rtype: dict
        :return: The short term storage request.
        """
        url = f"{self.gi.url}/short_term_storage/{storage_request_id}"
        is_ready_url = f"{url}/ready"

        def check_and_get_short_term_storage() -> Dict[str, Any]:
            if self._get(url=is_ready_url):
                return self._get(url=url)
            raise NotReady(f"Storage request {storage_request_id} is not ready")

        return wait_on(check_and_get_short_term_storage, maxwait=maxwait, interval=interval)

    def get_invocation_biocompute_object(self, invocation_id: str, maxwait: float = 1200) -> Dict[str, Any]:
        """
        Get a BioCompute object for an invocation.

        :type invocation_id: str
        :param invocation_id: Encoded workflow invocation ID

        :type maxwait: float
        :param maxwait: Total time (in seconds) to wait for the BioCompute
          object to become ready. After this time, a ``TimeoutException`` will
          be raised.

        :rtype: dict
        :return: The BioCompute object
        """
        url = self._make_url(invocation_id) + "/prepare_store_download"
        payload = {"model_store_format": "bco.json"}
        try:
            psd = self._post(url=url, payload=payload)
        except ConnectionError as e:
            if e.status_code not in (400, 404):
                raise
            # Galaxy release_22.05 and earlier
            url = self._make_url(invocation_id) + "/biocompute"
            return self._get(url=url)
        else:
            storage_request_id = psd["storage_request_id"]
            url = f"{self.gi.url}/short_term_storage/{storage_request_id}/ready"
            return self._wait_for_short_term_storage(storage_request_id, maxwait=maxwait)

    def wait_for_invocation(
        self, invocation_id: str, maxwait: float = 12000, interval: float = 3, check: bool = True
    ) -> Dict[str, Any]:
        """
        Wait until an invocation is in a terminal state.

        :type invocation_id: str
        :param invocation_id: Invocation ID to wait for.

        :type maxwait: float
        :param maxwait: Total time (in seconds) to wait for the invocation state
          to become terminal. After this time, a ``TimeoutException`` will be
          raised.

        :type interval: float
        :param interval: Time (in seconds) to wait between 2 consecutive checks.

        :type check: bool
        :param check: Whether to check if the invocation terminal state is
          'scheduled'.

        :rtype: dict
        :return: Details of the workflow invocation.
        """

        def check_and_get_invocation() -> Dict[str, Any]:
            invocation = self.gi.invocations.show_invocation(invocation_id)
            state = invocation["state"]
            if state in INVOCATION_TERMINAL_STATES:
                if check and state != "scheduled":
                    raise Exception(f"Invocation {invocation_id} is in terminal state {state}")
                return invocation
            raise NotReady(f"Invocation {invocation_id} is in non-terminal state {state}")

        return wait_on(check_and_get_invocation, maxwait=maxwait, interval=interval)

    def _invocation_step_url(self, invocation_id: str, step_id: str) -> str:
        return "/".join((self._make_url(invocation_id), "steps", step_id))


__all__ = ("InvocationClient",)



"""
Contains possible interactions with the Galaxy Roles
"""

from typing import (
    Any,
    Dict,
    List,
    Optional,
    TYPE_CHECKING,
)

from bioblend.galaxy.client import Client

if TYPE_CHECKING:
    from bioblend.galaxy import GalaxyInstance


class RolesClient(Client):
    module = "roles"

    def __init__(self, galaxy_instance: "GalaxyInstance") -> None:
        super().__init__(galaxy_instance)

    def get_roles(self) -> List[Dict[str, Any]]:
        """
        Displays a collection (list) of roles.

        :rtype: list
        :return: A list of dicts with details on individual roles.
          For example::

            [{"id": "f2db41e1fa331b3e",
              "model_class": "Role",
              "name": "Foo",
              "url": "/api/roles/f2db41e1fa331b3e"},
             {"id": "f597429621d6eb2b",
              "model_class": "Role",
              "name": "Bar",
              "url": "/api/roles/f597429621d6eb2b"}]
        """
        return self._get()

    def show_role(self, role_id: str) -> Dict[str, Any]:
        """
        Display information on a single role

        :type role_id: str
        :param role_id: Encoded role ID

        :rtype: dict
        :return: Details of the given role.
          For example::

            {"description": "Private Role for Foo",
             "id": "f2db41e1fa331b3e",
             "model_class": "Role",
             "name": "Foo",
             "type": "private",
             "url": "/api/roles/f2db41e1fa331b3e"}
        """
        return self._get(id=role_id)

    def create_role(
        self,
        role_name: str,
        description: str,
        user_ids: Optional[List[str]] = None,
        group_ids: Optional[List[str]] = None,
    ) -> Dict[str, Any]:
        """
        Create a new role.

        :type role_name: str
        :param role_name: A name for the new role

        :type description: str
        :param description: Description for the new role

        :type user_ids: list
        :param user_ids: A list of encoded user IDs to add to the new role

        :type group_ids: list
        :param group_ids: A list of encoded group IDs to add to the new role

        :rtype: dict
        :return: Details of the newly created role.
          For example::

            {'description': 'desc',
             'url': '/api/roles/ebfb8f50c6abde6d',
             'model_class': 'Role',
             'type': 'admin',
             'id': 'ebfb8f50c6abde6d',
             'name': 'Foo'}

        .. versionchanged:: 0.15.0
            Changed the return value from a 1-element list to a dict.
        """
        if user_ids is None:
            user_ids = []
        if group_ids is None:
            group_ids = []
        payload = {"name": role_name, "description": description, "user_ids": user_ids, "group_ids": group_ids}
        ret = self._post(payload)
        if isinstance(ret, list):
            # Galaxy release_20.09 and earlier returned a 1-element list
            ret = ret[0]
        return ret



import os
from typing import (
    Any,
    IO,
    NamedTuple,
    Optional,
    Type,
    TypeVar,
)


class FileStream(NamedTuple):
    name: str
    fd: IO

    def close(self) -> None:
        self.fd.close()


def attach_file(path: str, name: Optional[str] = None) -> FileStream:
    """
    Attach a path to a request payload object.

    :type path: str
    :param path: Path to file to attach to payload.

    :type name: str
    :param name: Name to give file, if different than actual pathname.

    :rtype: object
    :return: Returns an object compatible with requests post operation and
             capable of being closed with a ``close()`` method.
    """
    if name is None:
        name = os.path.basename(path)
    return FileStream(name, open(path, "rb"))


T = TypeVar("T")


def abstractclass(decorated_cls: Type[T]) -> Type[T]:
    """
    Decorator that marks a class as abstract even without any abstract method

    Adapted from https://stackoverflow.com/a/49013561/4503125
    """

    def clsnew(cls: Type[T], *args: Any, **kwargs: Any) -> T:
        # assert issubclass(cls, decorated_cls)
        if cls is decorated_cls:
            cls_name = getattr(decorated_cls, "__name__", str(decorated_cls))
            raise TypeError(f"Can't instantiate abstract class {cls_name}")
        return super(decorated_cls, cls).__new__(cls)  # type: ignore[misc]

    decorated_cls.__new__ = clsnew  # type: ignore[assignment]
    return decorated_cls


__all__ = (
    "abstractclass",
    "attach_file",
)



import os
import unittest

import bioblend
import bioblend.toolshed
from . import test_util


@test_util.skip_unless_toolshed()
class TestToolshed(unittest.TestCase):
    def setUp(self):
        toolshed_url = os.environ["BIOBLEND_TOOLSHED_URL"]
        self.ts = bioblend.toolshed.ToolShedInstance(url=toolshed_url)

    def test_categories_client(self):
        # get_categories
        categories = self.ts.categories.get_categories()
        assert "Assembly" in [c["name"] for c in categories]
        # we cannot test get_categories with deleted=True as it requires administrator status

        # show_category
        visualization_category_id = [c for c in categories if c["name"] == "Visualization"][0]["id"]
        visualization_category = self.ts.categories.show_category(visualization_category_id)
        assert visualization_category["description"] == "Tools for visualizing data"

        # get_repositories
        repositories = self.ts.categories.get_repositories(visualization_category_id)
        repositories_reversed = self.ts.categories.get_repositories(visualization_category_id, sort_order="desc")
        assert len(repositories["repositories"]) > 200
        assert {
            "deprecated",
            "description",
            "homepage_url",
            "id",
            "name",
            "owner",
            "remote_repository_url",
            "type",
            "update_time",
        } <= set(repositories["repositories"][0].keys())
        assert repositories["repositories"][0] == repositories_reversed["repositories"][-1]

    def test_repositories_client(self):
        # get_repositories
        repositories = self.ts.repositories.get_repositories()
        assert len(repositories) > 5000
        repository0 = repositories[0]
        for key in ("id", "name", "owner", "type", "description", "deprecated"):
            assert key in repository0

        repositories = self.ts.repositories.get_repositories(name="bam_to_sam", owner="devteam")
        assert len(repositories) == 1
        bam_to_sam_repo = repositories[0]
        assert bam_to_sam_repo["name"] == "bam_to_sam"
        assert bam_to_sam_repo["owner"] == "devteam"
        assert bam_to_sam_repo["type"] == "unrestricted"
        assert not bam_to_sam_repo["deprecated"]

        # search_repositories
        samtools_search = self.ts.repositories.search_repositories("samtools", page_size=5)
        assert int(samtools_search["total_results"]) > 20
        assert len(samtools_search["hits"]) == 5

        # show_repository
        show_bam_to_sam_repo = self.ts.repositories.show_repository(bam_to_sam_repo["id"])
        assert "SAM" in show_bam_to_sam_repo["long_description"]

        # test_create_repository
        # need to provide an API key to test this

        # test_update_repository
        # need to provide an API key to test this

    def test_repositories_revisions(self):
        # get_ordered_installable_revisions
        bam_to_sam_revisions = self.ts.repositories.get_ordered_installable_revisions("bam_to_sam", "devteam")
        assert len(bam_to_sam_revisions) >= 4

        # get_repository_revision_install_info
        bam_to_sam_revision_install_info = self.ts.repositories.get_repository_revision_install_info(
            "bam_to_sam", "devteam", bam_to_sam_revisions[0]
        )
        assert len(bam_to_sam_revision_install_info) == 3
        assert bam_to_sam_revision_install_info[0].get("model_class") == "Repository"
        assert bam_to_sam_revision_install_info[1].get("model_class") == "RepositoryMetadata"
        assert bam_to_sam_revision_install_info[2].get("model_class") is None

    def test_tools_client(self):
        # search_tools
        samtools_search = self.ts.tools.search_tools("samtools", page_size=5)
        assert int(samtools_search["page"]) == 1
        assert len(samtools_search["hits"]) == 5
        hit0_tool = samtools_search["hits"][0]["tool"]
        for key in ("id", "repo_owner_username", "repo_name", "name", "description"):
            assert key in hit0_tool



"""
Contains possible interactions with the Galaxy FTP Files
"""

from typing import (
    List,
    TYPE_CHECKING,
)

from bioblend.galaxy.client import Client

if TYPE_CHECKING:
    from bioblend.galaxy import GalaxyInstance


class FTPFilesClient(Client):
    module = "ftp_files"

    def __init__(self, galaxy_instance: "GalaxyInstance") -> None:
        super().__init__(galaxy_instance)

    def get_ftp_files(self, deleted: bool = False) -> List[dict]:
        """
        Get a list of local files.

        :type deleted: bool
        :param deleted: Whether to include deleted files

        :rtype: list
        :return: A list of dicts with details on individual files on FTP
        """
        return self._get()



"""
Clients for interacting with specific Galaxy entity types.

Classes in this module should not be instantiated directly, but used
via their handles in :class:`~.galaxy_instance.GalaxyInstance`.
"""

import abc
import json
from collections.abc import Sequence
from typing import (
    Any,
    cast,
    Dict,
    Generic,
    List,
    Literal,
    Optional,
    overload,
    Type,
    TYPE_CHECKING,
    Union,
)

import bioblend
from bioblend.galaxy.datasets import HdaLdda
from . import wrappers

if TYPE_CHECKING:
    from .galaxy_instance import GalaxyInstance


class ObjClient(abc.ABC):
    def __init__(self, obj_gi: "GalaxyInstance") -> None:
        self.obj_gi = obj_gi
        self.gi = self.obj_gi.gi
        self.log = bioblend.log

    @abc.abstractmethod
    def get(self, id_: str) -> wrappers.Wrapper:
        """
        Retrieve the object corresponding to the given id.
        """

    @abc.abstractmethod
    def get_previews(self, **kwargs: Any) -> list:
        """
        Get a list of object previews.

        Previews entity summaries provided by REST collection URIs, e.g.
        ``http://host:port/api/libraries``.  Being the most lightweight objects
        associated to the various entities, these are the ones that should be
        used to retrieve their basic info.

        :rtype: list
        :return: a list of object previews
        """

    @abc.abstractmethod
    def list(self) -> list:
        """
        Get a list of objects.

        This method first gets the entity summaries, then gets the complete
        description for each entity with an additional GET call, so may be slow.

        :rtype: list
        :return: a list of objects
        """

    def _select_id(self, id_: Optional[str] = None, name: Optional[str] = None) -> str:
        """
        Return the id that corresponds to the given id or name info.
        """
        if id_ is None and name is None:
            raise ValueError("Neither id nor name provided")
        if id_ is not None and name is not None:
            raise ValueError("Both id and name provided")
        if id_ is None:
            id_list = [_.id for _ in self.get_previews(name=name)]
            if len(id_list) > 1:
                raise ValueError(f"Ambiguous name '{name}'")
            if not id_list:
                raise ValueError(f"Name '{name}' not found")
            return id_list[0]
        else:
            return id_

    def _get_dict(self, meth_name: str, reply: Optional[Union[Dict[str, Any], List[Dict[str, Any]]]]) -> Dict[str, Any]:
        if reply is None:
            raise RuntimeError(f"{meth_name}: no reply")
        elif isinstance(reply, dict):
            return reply
        try:
            return reply[0]
        except (TypeError, IndexError):
            raise RuntimeError(f"{meth_name}: unexpected reply: {reply!r}")


class ObjDatasetContainerClient(
    ObjClient, Generic[wrappers.DatasetContainerSubtype, wrappers.DatasetContainerPreviewSubtype]
):
    CONTAINER_TYPE: Type[wrappers.DatasetContainer]
    CONTAINER_PREVIEW_TYPE: Type[wrappers.DatasetContainerPreview]

    def __init__(self, obj_gi: "GalaxyInstance") -> None:
        super().__init__(obj_gi=obj_gi)
        gi_client = getattr(self.gi, self.CONTAINER_TYPE.API_MODULE)
        get_fname = f"get_{self.CONTAINER_TYPE.API_MODULE}"
        self._get_f = getattr(gi_client, get_fname)
        show_fname = f"show_{self.CONTAINER_TYPE.__name__.lower()}"
        self._show_f = getattr(gi_client, show_fname)

    def get_previews(
        self, name: Optional[str] = None, deleted: bool = False, **kwargs: Any
    ) -> List[wrappers.DatasetContainerPreviewSubtype]:
        dicts = self._get_f(name=name, deleted=deleted, **kwargs)
        return [
            cast(wrappers.DatasetContainerPreviewSubtype, self.CONTAINER_PREVIEW_TYPE(_, gi=self.obj_gi)) for _ in dicts
        ]

    def get(self, id_: str) -> wrappers.DatasetContainerSubtype:
        """
        Retrieve the dataset container corresponding to the given id.
        """
        cdict = self._show_f(id_)
        c_infos = self._show_f(id_, contents=True)
        if not isinstance(c_infos, Sequence):
            raise RuntimeError(f"{self._show_f.__name__}: unexpected reply: {c_infos!r}")
        c_infos = [self.CONTAINER_TYPE.CONTENT_INFO_TYPE(_) for _ in c_infos]
        return cast(wrappers.DatasetContainerSubtype, self.CONTAINER_TYPE(cdict, content_infos=c_infos, gi=self.obj_gi))


class ObjLibraryClient(ObjDatasetContainerClient[wrappers.Library, wrappers.LibraryPreview]):
    """
    Interacts with Galaxy libraries.
    """

    CONTAINER_TYPE = wrappers.Library
    CONTAINER_PREVIEW_TYPE = wrappers.LibraryPreview

    def create(self, name: str, description: Optional[str] = None, synopsis: Optional[str] = None) -> wrappers.Library:
        """
        Create a data library with the properties defined in the arguments.

        :rtype: :class:`~.wrappers.Library`
        :return: the library just created
        """
        res = self.gi.libraries.create_library(name, description, synopsis)
        lib_info = self._get_dict("create_library", res)
        return self.get(lib_info["id"])

    def list(self, name: Optional[str] = None, deleted: bool = False) -> List[wrappers.Library]:
        """
        Get libraries owned by the user of this Galaxy instance.

        :type name: str
        :param name: return only libraries with this name
        :type deleted: bool
        :param deleted: if ``True``, return libraries that have been deleted

        :rtype: list of :class:`~.wrappers.Library`
        """
        dicts = self.gi.libraries.get_libraries(name=name, deleted=deleted)
        if not deleted:
            # return Library objects only for not-deleted libraries since Galaxy
            # does not filter them out and Galaxy release_14.08 and earlier
            # crashes when trying to get a deleted library
            return [self.get(_["id"]) for _ in dicts if not _["deleted"]]
        else:
            return [self.get(_["id"]) for _ in dicts]

    def delete(self, id_: Optional[str] = None, name: Optional[str] = None) -> None:
        """
        Delete the library with the given id or name.

        Fails if multiple libraries have the specified name.

        .. warning::
          Deleting a data library is irreversible - all of the data from
          the library will be permanently deleted.
        """
        id_ = self._select_id(id_=id_, name=name)
        res = self.gi.libraries.delete_library(id_)
        if not isinstance(res, dict):
            raise RuntimeError(f"delete_library: unexpected reply: {res!r}")


class ObjHistoryClient(ObjDatasetContainerClient[wrappers.History, wrappers.HistoryPreview]):
    """
    Interacts with Galaxy histories.
    """

    CONTAINER_TYPE = wrappers.History
    CONTAINER_PREVIEW_TYPE = wrappers.HistoryPreview

    def create(self, name: Optional[str] = None) -> wrappers.History:
        """
        Create a new Galaxy history, optionally setting its name.

        :rtype: :class:`~.wrappers.History`
        :return: the history just created
        """
        res = self.gi.histories.create_history(name=name)
        hist_info = self._get_dict("create_history", res)
        return self.get(hist_info["id"])

    def list(self, name: Optional[str] = None, deleted: bool = False) -> List[wrappers.History]:
        """
        Get histories owned by the user of this Galaxy instance.

        :type name: str
        :param name: return only histories with this name
        :type deleted: bool
        :param deleted: if ``True``, return histories that have been deleted

        :rtype: list of :class:`~.wrappers.History`
        """
        dicts = self.gi.histories.get_histories(name=name, deleted=deleted)
        return [self.get(_["id"]) for _ in dicts]

    def delete(self, id_: Optional[str] = None, name: Optional[str] = None, purge: bool = False) -> None:
        """
        Delete the history with the given id or name.

        Fails if multiple histories have the same name.

        :type purge: bool
        :param purge: if ``True``, also purge (permanently delete) the history

        .. note::
          The ``purge`` option works only if the Galaxy instance has the
          ``allow_user_dataset_purge`` option set to ``true`` in the
          ``config/galaxy.yml`` configuration file.
        """
        id_ = self._select_id(id_=id_, name=name)
        res = self.gi.histories.delete_history(id_, purge=purge)
        if not isinstance(res, dict):
            raise RuntimeError(f"delete_history: unexpected reply: {res!r}")


class ObjWorkflowClient(ObjClient):
    """
    Interacts with Galaxy workflows.
    """

    def import_new(self, src: Union[str, Dict[str, Any]], publish: bool = False) -> wrappers.Workflow:
        """
        Imports a new workflow into Galaxy.

        :type src: dict or str
        :param src: deserialized (dictionary) or serialized (str) JSON
          dump of the workflow (this is normally obtained by exporting
          a workflow from Galaxy).

        :type publish: bool
        :param publish:  if ``True`` the uploaded workflow will be published;
                         otherwise it will be visible only by the user which uploads it (default).

        :rtype: :class:`~.wrappers.Workflow`
        :return: the workflow just imported
        """
        if isinstance(src, dict):
            wf_dict = src
        else:
            try:
                wf_dict = json.loads(src)
            except (TypeError, ValueError):
                raise ValueError(f"src not supported: {src!r}")
        wf_info = self.gi.workflows.import_workflow_dict(wf_dict, publish)
        return self.get(wf_info["id"])

    def import_shared(self, id_: str) -> wrappers.Workflow:
        """
        Imports a shared workflow to the user's space.

        :type id_: str
        :param id_: workflow id

        :rtype: :class:`~.wrappers.Workflow`
        :return: the workflow just imported
        """
        wf_info = self.gi.workflows.import_shared_workflow(id_)
        return self.get(wf_info["id"])

    def get(self, id_: str) -> wrappers.Workflow:
        """
        Retrieve the workflow corresponding to the given id.

        :rtype: :class:`~.wrappers.Workflow`
        :return: the workflow corresponding to ``id_``
        """
        res = self.gi.workflows.show_workflow(id_)
        wf_dict = self._get_dict("show_workflow", res)
        return wrappers.Workflow(wf_dict, gi=self.obj_gi)

    # the 'deleted' option is not available for workflows
    def get_previews(
        self, name: Optional[str] = None, published: bool = False, **kwargs: Any
    ) -> List[wrappers.WorkflowPreview]:
        dicts = self.gi.workflows.get_workflows(name=name, published=published, **kwargs)
        return [wrappers.WorkflowPreview(_, gi=self.obj_gi) for _ in dicts]

    # the 'deleted' option is not available for workflows
    def list(self, name: Optional[str] = None, published: bool = False) -> List[wrappers.Workflow]:
        """
        Get workflows owned by the user of this Galaxy instance.

        :type name: str
        :param name: return only workflows with this name
        :type published: bool
        :param published: if ``True``, return also published workflows

        :rtype: list of :class:`~.wrappers.Workflow`
        """
        dicts = self.gi.workflows.get_workflows(name=name, published=published)
        return [self.get(_["id"]) for _ in dicts]

    def delete(self, id_: Optional[str] = None, name: Optional[str] = None) -> None:
        """
        Delete the workflow with the given id or name.

        Fails if multiple workflows have the specified name.

        .. warning::
          Deleting a workflow is irreversible - all of the data from
          the workflow will be permanently deleted.
        """
        id_ = self._select_id(id_=id_, name=name)
        self.gi.workflows.delete_workflow(id_)


class ObjInvocationClient(ObjClient):
    """
    Interacts with Galaxy Invocations.
    """

    def get(self, id_: str) -> wrappers.Invocation:
        """
        Get an invocation by ID.

        :rtype: Invocation
        :param: invocation object
        """
        inv_dict = self.gi.invocations.show_invocation(id_)
        return wrappers.Invocation(inv_dict, self.obj_gi)

    def get_previews(self, **kwargs: Any) -> List[wrappers.InvocationPreview]:
        """
        Get previews of all invocations.

        :rtype: list of InvocationPreview
        :param: previews of invocations
        """
        inv_list = self.gi.invocations.get_invocations(**kwargs)
        return [wrappers.InvocationPreview(inv_dict, gi=self.obj_gi) for inv_dict in inv_list]

    def list(
        self,
        workflow: Optional[wrappers.Workflow] = None,
        history: Optional[wrappers.History] = None,
        include_terminal: bool = True,
        limit: Optional[int] = None,
    ) -> List[wrappers.Invocation]:
        """
        Get full listing of workflow invocations, or select a subset
        by specifying optional arguments for filtering (e.g. a workflow).

        :type workflow: wrappers.Workflow
        :param workflow: Include only invocations associated with
          this workflow

        :type history: wrappers.History
        :param history: Include only invocations associated with
          this history

        :param include_terminal: bool
        :param: Whether to include invocations in terminal states

        :type limit: int
        :param limit: Maximum number of invocations to return - if specified,
          the most recent invocations will be returned.

        :rtype: list of Invocation
        :param: invocation objects
        """
        inv_dict_list = self.gi.invocations.get_invocations(
            workflow_id=workflow.id if workflow else None,
            history_id=history.id if history else None,
            include_terminal=include_terminal,
            limit=limit,
            view="element",
            step_details=True,
        )
        return [wrappers.Invocation(inv_dict, self.obj_gi) for inv_dict in inv_dict_list]


class ObjToolClient(ObjClient):
    """
    Interacts with Galaxy tools.
    """

    def get(self, id_: str, io_details: bool = False, link_details: bool = False) -> wrappers.Tool:
        """
        Retrieve the tool corresponding to the given id.

        :type io_details: bool
        :param io_details: if True, get also input and output details

        :type link_details: bool
        :param link_details: if True, get also link details

        :rtype: :class:`~.wrappers.Tool`
        :return: the tool corresponding to ``id_``
        """
        res = self.gi.tools.show_tool(id_, io_details=io_details, link_details=link_details)
        tool_dict = self._get_dict("show_tool", res)
        return wrappers.Tool(tool_dict, gi=self.obj_gi)

    def get_previews(self, name: Optional[str] = None, trackster: bool = False, **kwargs: Any) -> List[wrappers.Tool]:
        """
        Get the list of tools installed on the Galaxy instance.

        :type name: str
        :param name: return only tools with this name

        :type trackster: bool
        :param trackster: if True, only tools that are compatible with
          Trackster are returned

        :rtype: list of :class:`~.wrappers.Tool`
        """
        dicts = self.gi.tools.get_tools(name=name, trackster=trackster, **kwargs)
        return [wrappers.Tool(_, gi=self.obj_gi) for _ in dicts]

    # the 'deleted' option is not available for tools
    def list(self, name: Optional[str] = None, trackster: bool = False) -> List[wrappers.Tool]:
        """
        Get the list of tools installed on the Galaxy instance.

        :type name: str
        :param name: return only tools with this name

        :type trackster: bool
        :param trackster: if True, only tools that are compatible with
          Trackster are returned

        :rtype: list of :class:`~.wrappers.Tool`
        """
        # dicts = self.gi.tools.get_tools(name=name, trackster=trackster)
        # return [self.get(_['id']) for _ in dicts]
        # As of 2015/04/15, GET /api/tools returns also data manager tools for
        # non-admin users, see
        # https://trello.com/c/jyl0cvFP/2633-api-tool-list-filtering-doesn-t-filter-data-managers-for-non-admins
        # Trying to get() a data manager tool would then return a 404 Not Found
        # error.
        # Moreover, the dicts returned by gi.tools.get_tools() are richer than
        # those returned by get(), so make this an alias for get_previews().
        return self.get_previews(name, trackster)


class ObjJobClient(ObjClient):
    """
    Interacts with Galaxy jobs.
    """

    def get(self, id_: str, full_details: bool = False) -> wrappers.Job:
        """
        Retrieve the job corresponding to the given id.

        :type full_details: bool
        :param full_details: if ``True``, return the complete list of details
          for the given job.

        :rtype: :class:`~.wrappers.Job`
        :return: the job corresponding to ``id_``
        """
        res = self.gi.jobs.show_job(id_, full_details)
        job_dict = self._get_dict("show_job", res)
        return wrappers.Job(job_dict, gi=self.obj_gi)

    def get_previews(self, **kwargs: Any) -> List[wrappers.JobPreview]:
        dicts = self.gi.jobs.get_jobs(**kwargs)
        return [wrappers.JobPreview(_, gi=self.obj_gi) for _ in dicts]

    def list(self) -> List[wrappers.Job]:
        """
        Get the list of jobs of the current user.

        :rtype: list of :class:`~.wrappers.Job`
        """
        dicts = self.gi.jobs.get_jobs()
        return [self.get(_["id"]) for _ in dicts]


class ObjDatasetClient(ObjClient):
    """
    Interacts with Galaxy datasets.
    """

    @overload
    def get(self, id_: str, hda_ldda: Literal["hda"] = "hda") -> wrappers.HistoryDatasetAssociation: ...

    @overload
    def get(self, id_: str, hda_ldda: Literal["ldda"]) -> wrappers.LibraryDatasetDatasetAssociation: ...

    def get(self, id_: str, hda_ldda: HdaLdda = "hda") -> wrappers.Dataset:
        """
        Retrieve the dataset corresponding to the given id.

        :type hda_ldda: str
        :param hda_ldda: Whether to show a history dataset ('hda' - the default)
          or library dataset ('ldda')

        :rtype: :class:`~.wrappers.HistoryDatasetAssociation` or :class:`~.wrappers.LibraryDatasetDatasetAssociation`
        :return: the history or library dataset corresponding to ``id_``
        """
        res = self.gi.datasets.show_dataset(id_, hda_ldda=hda_ldda)
        ds_dict = self._get_dict("show_dataset", res)
        if hda_ldda == "hda":
            hist = self.obj_gi.histories.get(ds_dict["history_id"])
            return wrappers.HistoryDatasetAssociation(ds_dict, hist, gi=self.obj_gi)
        elif hda_ldda == "ldda":
            lib = self.obj_gi.libraries.get(ds_dict["parent_library_id"])
            return wrappers.LibraryDatasetDatasetAssociation(ds_dict, lib, gi=self.obj_gi)
        else:
            raise ValueError(f"Unsupported value for hda_ldda: {hda_ldda}")

    def get_previews(self, **kwargs: Any) -> list:
        raise NotImplementedError()

    def list(self) -> list:
        raise NotImplementedError()


class ObjDatasetCollectionClient(ObjClient):
    """
    Interacts with Galaxy dataset collections.
    """

    def get(self, id_: str) -> wrappers.HistoryDatasetCollectionAssociation:
        """
        Retrieve the dataset collection corresponding to the given id.

        :rtype: :class:`~.wrappers.HistoryDatasetCollectionAssociation`
        :return: the history dataset collection corresponding to ``id_``
        """
        res = self.gi.dataset_collections.show_dataset_collection(id_)
        ds_dict = self._get_dict("show_dataset_collection", res)
        hist = self.obj_gi.histories.get(ds_dict["history_id"])
        return wrappers.HistoryDatasetCollectionAssociation(ds_dict, hist, gi=self.obj_gi)

    def get_previews(self, **kwargs: Any) -> list:
        raise NotImplementedError()

    def list(self) -> list:
        raise NotImplementedError()



"""
This example retrieves details of all the Data Libraries available to us and lists information on them.

Usage: python list_data_libraries.py <Galaxy_URL> <Galaxy_API_key>
"""

import sys

from bioblend.galaxy.objects import GalaxyInstance

if len(sys.argv) != 3:
    print("Usage: python list_data_libraries.py <Galaxy_URL> <Galaxy_API_key>")
    sys.exit(1)
galaxy_url = sys.argv[1]
galaxy_key = sys.argv[2]

print("Initiating Galaxy connection")

gi = GalaxyInstance(galaxy_url, api_key=galaxy_key)

print("Retrieving Data Library list")

libraries = gi.libraries.get_previews()

if len(libraries) == 0:
    print("There are no Data Libraries available.")
else:
    print("\nData Libraries:")
    for lib in libraries:
        print(f"{lib.name} : {lib.id}")



from typing import (
    Any,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
)


class InputsBuilder:
    """ """

    def __init__(self) -> None:
        self._input_dict: Dict[str, Any] = {}

    def set(self, name: str, input: Any) -> "InputsBuilder":
        self._input_dict[name] = input
        return self

    def set_param(self, name: str, value: Any) -> "InputsBuilder":
        return self.set(name, param(value=value))

    def set_dataset_param(self, name: str, value: str, src: str = "hda") -> "InputsBuilder":
        return self.set(name, dataset(value, src=src))

    def to_dict(self) -> Dict[str, Any]:
        values = {}
        for key, value in self.flat_iter():
            if hasattr(value, "value"):
                value = value.value
            values[key] = value
        return values

    def flat_iter(self, prefix: Optional[str] = None) -> Iterator[Tuple[str, Any]]:
        for key, value in self._input_dict.items():
            effective_key = key if prefix is None else f"{prefix}|{key}"
            if hasattr(value, "flat_iter"):
                yield from value.flat_iter(effective_key)
            else:
                yield effective_key, value


class RepeatBuilder:
    def __init__(self) -> None:
        self._instances: List[InputsBuilder] = []

    def instance(self, inputs: InputsBuilder) -> "RepeatBuilder":
        self._instances.append(inputs)
        return self

    def flat_iter(self, prefix: str) -> Iterator[Tuple[str, Any]]:
        for index, instance in enumerate(self._instances):
            index_prefix = f"{prefix}_{index}"
            yield from instance.flat_iter(index_prefix)


class Param:
    def __init__(self, value: Any) -> None:
        self.value = value


class DatasetParam(Param):
    def __init__(self, value: Union[Dict[str, str], str], src: str = "hda") -> None:
        if not isinstance(value, dict):
            value = {"src": src, "id": value}
        super().__init__(value)


inputs = InputsBuilder
repeat = RepeatBuilder
conditional = InputsBuilder
param = Param
dataset = DatasetParam

__all__ = ("inputs", "repeat", "conditional", "param")



"""
Contains possible interactions with the Galaxy Jobs
"""

import logging
from typing import (
    Any,
    Dict,
    List,
    Literal,
    Optional,
    TYPE_CHECKING,
)

from bioblend import (
    NotReady,
    wait_on,
)
from bioblend.galaxy.client import Client

if TYPE_CHECKING:
    from bioblend.galaxy import GalaxyInstance

log = logging.getLogger(__name__)

JOB_TERMINAL_STATES = {"deleted", "deleting", "error", "ok"}
# Job non-terminal states are: 'deleted_new', 'failed', 'new', 'paused',
# 'queued', 'resubmitted', 'running', 'upload', 'waiting'


class JobsClient(Client):
    module = "jobs"

    def __init__(self, galaxy_instance: "GalaxyInstance") -> None:
        super().__init__(galaxy_instance)

    def get_jobs(
        self,
        state: Optional[str] = None,
        history_id: Optional[str] = None,
        invocation_id: Optional[str] = None,
        tool_id: Optional[str] = None,
        workflow_id: Optional[str] = None,
        user_id: Optional[str] = None,
        date_range_min: Optional[str] = None,
        date_range_max: Optional[str] = None,
        limit: int = 500,
        offset: int = 0,
        user_details: bool = False,
        order_by: Literal["create_time", "update_time"] = "update_time",
    ) -> List[Dict[str, Any]]:
        """
        Get all jobs, or select a subset by specifying optional arguments for
        filtering (e.g. a state).

        If the user is an admin, this will return jobs for all the users,
        otherwise only for the current user.

        :type state: str or list of str
        :param state: Job states to filter on.

        :type history_id: str
        :param history_id: Encoded history ID to filter on.

        :type invocation_id: string
        :param invocation_id: Encoded workflow invocation ID to filter on.

        :type tool_id: str or list of str
        :param tool_id: Tool IDs to filter on.

        :type workflow_id: string
        :param workflow_id: Encoded workflow ID to filter on.

        :type user_id: str
        :param user_id: Encoded user ID to filter on. Only admin users can
          access the jobs of other users.

        :type date_range_min: str
        :param date_range_min: Mininum job update date (in YYYY-MM-DD format) to
          filter on.

        :type date_range_max: str
        :param date_range_max: Maximum job update date (in YYYY-MM-DD format) to
          filter on.

        :type limit: int
        :param limit: Maximum number of jobs to return.

        :type offset: int
        :param offset: Return jobs starting from this specified position.
          For example, if ``limit`` is set to 100 and ``offset`` to 200,
          jobs 200-299 will be returned.

        :type user_details: bool
        :param user_details: If ``True`` and the user is an admin, add the user
          email to each returned job dictionary.

        :type order_by: str
        :param order_by: Whether to order jobs by ``create_time`` or
          ``update_time`` (the default).

        :rtype: list of dict
        :return: Summary information for each selected job.
          For example::

            [{'create_time': '2014-03-01T16:16:48.640550',
              'exit_code': 0,
              'id': 'ebfb8f50c6abde6d',
              'model_class': 'Job',
              'state': 'ok',
              'tool_id': 'fasta2tab',
              'update_time': '2014-03-01T16:16:50.657399'},
             {'create_time': '2014-03-01T16:05:34.851246',
              'exit_code': 0,
              'id': '1cd8e2f6b131e891',
              'model_class': 'Job',
              'state': 'ok',
              'tool_id': 'upload1',
              'update_time': '2014-03-01T16:05:39.558458'}]

        .. note::
          The following parameters work only on Galaxy 21.05 or later: ``user_id``,
          ``limit``, ``offset``, ``workflow_id``, ``invocation_id``.
        """
        params: Dict[str, Any] = {"limit": limit, "offset": offset}
        if state:
            params["state"] = state
        if history_id:
            params["history_id"] = history_id
        if invocation_id:
            params["invocation_id"] = invocation_id
        if tool_id:
            params["tool_id"] = tool_id
        if workflow_id:
            params["workflow_id"] = workflow_id
        if user_id:
            params["user_id"] = user_id
        if date_range_min:
            params["date_range_min"] = date_range_min
        if date_range_max:
            params["date_range_max"] = date_range_max
        if user_details:
            params["user_details"] = user_details
        if order_by:
            params["order_by"] = order_by
        return self._get(params=params)

    def show_job(self, job_id: str, full_details: bool = False) -> Dict[str, Any]:
        """
        Get details of a given job of the current user.

        :type job_id: str
        :param job_id: job ID

        :type full_details: bool
        :param full_details: when ``True``, the complete list of details for the
          given job.

        :rtype: dict
        :return: A description of the given job.
          For example::

            {'create_time': '2014-03-01T16:17:29.828624',
             'exit_code': 0,
             'id': 'a799d38679e985db',
             'inputs': {'input': {'id': 'ebfb8f50c6abde6d', 'src': 'hda'}},
             'model_class': 'Job',
             'outputs': {'output': {'id': 'a799d38679e985db', 'src': 'hda'}},
             'params': {'chromInfo': '"/opt/galaxy-central/tool-data/shared/ucsc/chrom/?.len"',
                        'dbkey': '"?"',
                        'seq_col': '"2"',
                        'title_col': '["1"]'},
             'state': 'ok',
             'tool_id': 'tab2fasta',
             'update_time': '2014-03-01T16:17:31.930728'}
        """
        params = {}
        if full_details:
            params["full"] = full_details

        return self._get(id=job_id, params=params)

    def _build_for_rerun(self, job_id: str) -> Dict[str, Any]:
        """
        Get details of a given job that can be used to rerun the corresponding tool.

        :type job_id: str
        :param job_id: job ID

        :rtype: dict
        :return: A description of the given job, with all parameters required to rerun.

        """
        url = "/".join((self._make_url(job_id), "build_for_rerun"))
        return self._get(url=url)

    def rerun_job(
        self,
        job_id: str,
        remap: bool = False,
        tool_inputs_update: Optional[Dict[str, Any]] = None,
        history_id: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Rerun a job.

        :type job_id: str
        :param job_id: job ID

        :type remap: bool
        :param remap: when ``True``, the job output(s) will be remapped onto the dataset(s)
          created by the original job; if other jobs were waiting for this job to finish
          successfully, they will be resumed using the new outputs of this tool run. When
          ``False``, new job output(s) will be created. Note that if Galaxy does not permit
          remapping for the job in question, specifying ``True`` will result in an error.

        :type tool_inputs_update: dict
        :param tool_inputs_update: dictionary specifying any changes which should be
          made to tool parameters for the rerun job. This dictionary should have the same
          structure as is required when submitting the ``tool_inputs`` dictionary to
          ``gi.tools.run_tool()``, but only needs to include the inputs or parameters
          to be updated for the rerun job.

        :type history_id: str
        :param history_id: ID of the history in which the job should be executed; if
          not specified, the same history will be used as the original job run.

        :rtype: dict
        :return: Information about outputs and the rerun job

        .. note::
          This method works only on Galaxy 21.01 or later.
        """
        job_rerun_params = self._build_for_rerun(job_id)
        job_inputs = job_rerun_params["state_inputs"]

        if remap:
            if not job_rerun_params["job_remap"]:
                raise ValueError("remap was set to True, but this job is not remappable.")
            job_inputs["rerun_remap_job_id"] = job_id

        def update_inputs(inputs: Dict[str, Any], tool_inputs_update: Dict[str, Any]) -> None:
            """Recursively update inputs with tool_inputs_update"""
            for input_param, input_value in tool_inputs_update.items():
                if isinstance(input_value, dict):
                    update_inputs(inputs[input_param], input_value)
                else:
                    inputs[input_param] = input_value

        if tool_inputs_update:
            update_inputs(job_inputs, tool_inputs_update)

        url = "/".join((self.gi.url, "tools"))
        payload = {
            "history_id": history_id if history_id else job_rerun_params["history_id"],
            "tool_id": job_rerun_params["id"],
            "inputs": job_inputs,
            "input_format": "21.01",
        }
        return self._post(url=url, payload=payload)

    def get_state(self, job_id: str) -> str:
        """
        Display the current state for a given job of the current user.

        :type job_id: str
        :param job_id: job ID

        :rtype: str
        :return: state of the given job among the following values: `new`,
          `queued`, `running`, `waiting`, `ok`. If the state cannot be
          retrieved, an empty string is returned.

        .. versionadded:: 0.5.3
        """
        return self.show_job(job_id).get("state", "")

    def search_jobs(self, tool_id: str, inputs: Dict[str, Any], state: Optional[str] = None) -> List[Dict[str, Any]]:
        """
        Return jobs matching input parameters.

        :type tool_id: str
        :param tool_id: only return jobs associated with this tool ID

        :type inputs: dict
        :param inputs: return only jobs that have matching inputs

        :type state: str
        :param state: only return jobs in this state

        :rtype: list of dicts
        :return: Summary information for each matching job

        This method is designed to scan the list of previously run jobs and find
        records of jobs with identical input parameters and datasets. This can
        be used to minimize the amount of repeated work by simply recycling the
        old results.

        .. versionchanged:: 0.16.0
          Replaced the ``job_info`` parameter with separate ``tool_id``,
          ``inputs`` and ``state``.
        """
        job_info = {
            "tool_id": tool_id,
            "inputs": inputs,
        }
        if state:
            job_info["state"] = state
        url = self._make_url() + "/search"
        return self._post(url=url, payload=job_info)

    def get_metrics(self, job_id: str) -> List[Dict[str, Any]]:
        """
        Return job metrics for a given job.

        :type job_id: str
        :param job_id: job ID

        :rtype: list
        :return: list containing job metrics

        .. note::
          Calling ``show_job()`` with ``full_details=True`` also returns the
          metrics for a job if the user is an admin. This method allows to fetch
          metrics even as a normal user as long as the Galaxy instance has the
          ``expose_potentially_sensitive_job_metrics`` option set to ``true`` in
          the ``config/galaxy.yml`` configuration file.
        """
        url = self._make_url(module_id=job_id) + "/metrics"
        return self._get(url=url)

    def cancel_job(self, job_id: str) -> bool:
        """
        Cancel a job, deleting output datasets.

        :type job_id: str
        :param job_id: job ID

        :rtype: bool
        :return: ``True`` if the job was successfully cancelled, ``False`` if
          it was already in a terminal state before the cancellation.
        """
        return self._delete(id=job_id)

    def report_error(self, job_id: str, dataset_id: str, message: str, email: Optional[str] = None) -> Dict[str, Any]:
        """
        Report an error for a given job and dataset to the server administrators.

        :type job_id: str
        :param job_id: job ID

        :type dataset_id: str
        :param dataset_id: Dataset ID

        :type message: str
        :param message: Error message

        :type email: str
        :param email: Email for error report submission. If not specified, the email
          associated with the Galaxy user account is used by default.

        :rtype: dict
        :return: dict containing job error reply

        .. note::
          This method works only on Galaxy 20.01 or later.
        """
        payload = {
            "message": message,
            "dataset_id": dataset_id,
        }
        if email is not None:
            payload["email"] = email

        url = self._make_url(module_id=job_id) + "/error"
        return self._post(url=url, payload=payload)

    def get_common_problems(self, job_id: str) -> Dict[str, Any]:
        """
        Query inputs and jobs for common potential problems that might
        have resulted in job failure.

        :type job_id: str
        :param job_id: job ID

        :rtype: dict
        :return: dict containing potential problems

        .. note::
          This method works only on Galaxy 19.05 or later.
        """
        url = self._make_url(module_id=job_id) + "/common_problems"
        return self._get(url=url)

    def get_inputs(self, job_id: str) -> List[Dict[str, Any]]:
        """
        Get dataset inputs used by a job.

        :type job_id: str
        :param job_id: job ID

        :rtype: list of dicts
        :return: Inputs for the given job
        """
        url = self._make_url(module_id=job_id) + "/inputs"
        return self._get(url=url)

    def get_outputs(self, job_id: str) -> List[Dict[str, Any]]:
        """
        Get dataset outputs produced by a job.

        :type job_id: str
        :param job_id: job ID

        :rtype: list of dicts
        :return: Outputs of the given job
        """
        url = self._make_url(module_id=job_id) + "/outputs"
        return self._get(url=url)

    def resume_job(self, job_id: str) -> List[Dict[str, Any]]:
        """
        Resume a job if it is paused.

        :type job_id: str
        :param job_id: job ID

        :rtype: list of dicts
        :return: list of dictionaries containing output dataset associations
        """
        url = self._make_url(module_id=job_id) + "/resume"
        return self._put(url=url)

    def get_destination_params(self, job_id: str) -> Dict[str, Any]:
        """
        Get destination parameters for a job, describing
        the environment and location where the job is run.

        :type job_id: str
        :param job_id: job ID

        :rtype: dict
        :return: Destination parameters for the given job

        .. note::
          This method works only on Galaxy 20.05 or later and if the user is a
          Galaxy admin.
        """
        url = self._make_url(module_id=job_id) + "/destination_params"
        return self._get(url=url)

    def show_job_lock(self) -> bool:
        """
        Show whether the job lock is active or not. If it is active,
        no jobs will dispatch on the Galaxy server.

        :rtype: bool
        :return: Status of the job lock

        .. note::
          This method works only on Galaxy 20.05 or later and if the user is a
          Galaxy admin.
        """
        url = self.gi.url + "/job_lock"
        response = self._get(url=url)
        return response["active"]

    def update_job_lock(self, active: bool = False) -> bool:
        """
        Update the job lock status by setting ``active`` to either
        ``True`` or ``False``. If ``True``, all job dispatching will
        be blocked.

        :rtype: bool
        :return: Updated status of the job lock

        .. note::
          This method works only on Galaxy 20.05 or later and if the user is a
          Galaxy admin.
        """
        payload = {
            "active": active,
        }
        url = self.gi.url + "/job_lock"
        response = self._put(url=url, payload=payload)
        return response["active"]

    def wait_for_job(
        self, job_id: str, maxwait: float = 12000, interval: float = 3, check: bool = True
    ) -> Dict[str, Any]:
        """
        Wait until a job is in a terminal state.

        :type job_id: str
        :param job_id: job ID

        :type maxwait: float
        :param maxwait: Total time (in seconds) to wait for the job state to
          become terminal. After this time, a ``TimeoutException`` will be
          raised.

        :type interval: float
        :param interval: Time (in seconds) to wait between 2 consecutive checks.

        :type check: bool
        :param check: Whether to check if the job terminal state is 'ok'.

        :rtype: dict
        :return: Details of the given job.
        """

        def check_and_get_job() -> Dict[str, Any]:
            job = self.show_job(job_id)
            state = job["state"]
            if state in JOB_TERMINAL_STATES:
                if check and state != "ok":
                    raise Exception(f"Job {job_id} is in terminal state {state}")
                return job
            raise NotReady(f"Job {job_id} is in non-terminal state {state}")

        return wait_on(check_and_get_job, maxwait=maxwait, interval=interval)



"""
Tests on the GalaxyInstance object itself.
"""

import os
import time
import unittest

import pytest

from bioblend import ConnectionError
from bioblend.galaxy import GalaxyInstance
from . import test_util


class TestGalaxyInstance(unittest.TestCase):
    def setUp(self):
        # "connect" to a fake Galaxy instance
        self.gi = GalaxyInstance("http://localhost:56789", key="whatever")

    def test_url_attribute(self):
        assert self.gi.base_url == "http://localhost:56789"
        assert self.gi.url == "http://localhost:56789/api"
        # Test instance served at a subdirectory
        gi = GalaxyInstance("http://localhost:56789/galaxy/", key="whatever")
        assert gi.base_url == "http://localhost:56789/galaxy"
        assert gi.url == "http://localhost:56789/galaxy/api"

    def test_set_max_get_attempts(self):
        self.gi.max_get_attempts = 3
        assert 3 == self.gi.max_get_attempts

    def test_set_retry_delay(self):
        self.gi.get_retry_delay = 5.0
        assert 5.0 == self.gi.get_retry_delay

    def test_get_retry(self):
        # We set the client to try twice, with a delay of 5 seconds between
        # attempts. So, we expect the call to take at least 5 seconds before
        # failing.
        self.gi.max_get_attempts = 3
        self.gi.get_retry_delay = 2
        start = time.time()
        with pytest.raises(ConnectionError):
            self.gi.libraries.get_libraries()
        end = time.time()
        duration = end - start
        assert duration > self.gi.get_retry_delay * (self.gi.max_get_attempts - 1), "Didn't seem to retry long enough"

    def test_missing_scheme_fake_url(self):
        with pytest.raises(ValueError):
            GalaxyInstance("localhost:56789", key="whatever")

    @test_util.skip_unless_galaxy()
    def test_missing_scheme_real_url(self):
        galaxy_url = os.environ["BIOBLEND_GALAXY_URL"]
        # Strip the scheme from galaxy_url
        scheme_sep = "://"
        if scheme_sep in galaxy_url:
            galaxy_url = galaxy_url.partition(scheme_sep)[2]
        GalaxyInstance(url=galaxy_url)



"""
An interface the clients should implement.

This class is primarily a helper for the library and user code
should not use it directly.
"""

import time
from typing import (
    Any,
    Optional,
    TYPE_CHECKING,
)

import requests

import bioblend

# The following import must be preserved for compatibility because
# ConnectionError class was originally defined here
from bioblend import ConnectionError  # noqa: I202

if TYPE_CHECKING:
    from bioblend.galaxyclient import GalaxyClient


class Client:
    # The `module` attribute needs to be defined in subclasses
    module: str
    gi: "GalaxyClient"

    @classmethod
    def max_get_retries(cls) -> None:
        raise AttributeError("Deprecated method, please use gi's `max_get_attempts` property")

    @classmethod
    def set_max_get_retries(cls, value: int) -> None:
        raise AttributeError("Deprecated method, please use gi's `max_get_attempts` property")

    @classmethod
    def get_retry_delay(cls) -> None:
        raise AttributeError("Deprecated method, please use gi's `get_retry_delay` property")

    @classmethod
    def set_get_retry_delay(cls, value: float) -> None:
        raise AttributeError("Deprecated method, please use gi's `get_retry_delay` property")

    def __init__(self, galaxy_instance: "GalaxyClient") -> None:
        """
        A generic Client interface defining the common fields.

        All clients *must* define the following field (which will be
        used as part of the URL composition (e.g.,
        ``http://<galaxy_instance>/api/libraries``): ``self.module =
        'workflows' | 'libraries' | 'histories' | ...``
        """
        self.gi = galaxy_instance

    def _make_url(self, module_id: Optional[str] = None, deleted: bool = False, contents: bool = False) -> str:
        """
        Compose a URL based on the provided arguments.

        :type module_id: str
        :param module_id: The encoded ID for a specific module (eg, library ID)

        :type deleted: bool
        :param deleted: If ``True``, include ``deleted`` in the URL, after the module
                        name (eg, ``<base_url>/api/libraries/deleted``)

        :type contents: bool
        :param contents: If ``True``, include 'contents' in the URL, after the module ID:
                         ``<base_url>/api/libraries/<encoded_library_id>/contents``
        """
        c_url = "/".join((self.gi.url, self.module))
        if deleted:
            c_url = c_url + "/deleted"
        if module_id:
            c_url = "/".join((c_url, module_id))
            if contents:
                c_url = c_url + "/contents"
        return c_url

    def _get(
        self,
        id: Optional[str] = None,
        deleted: bool = False,
        contents: bool = False,
        url: Optional[str] = None,
        params: Optional[dict] = None,
        json: bool = True,
    ) -> Any:
        """
        Do a GET request, composing the URL from ``id``, ``deleted`` and
        ``contents``.  Alternatively, an explicit ``url`` can be provided.
        If ``json`` is set to ``True``, return a decoded JSON object
        (and treat an empty or undecodable response as an error).

        The request will optionally be retried as configured by gi's
        ``max_get_attempts`` and ``get_retry_delay``: this offers some
        resilience in the presence of temporary failures.

        :return: The decoded response if ``json`` is set to ``True``, otherwise
          the response object
        """
        if url is None:
            url = self._make_url(module_id=id, deleted=deleted, contents=contents)
        attempts_left = self.gi.max_get_attempts
        retry_delay = self.gi.get_retry_delay
        bioblend.log.debug("GET - attempts left: %s; retry delay: %s", attempts_left, retry_delay)
        msg = ""
        while attempts_left > 0:
            attempts_left -= 1
            try:
                r = self.gi.make_get_request(url, params=params)
            except requests.exceptions.ConnectionError as e:
                msg = str(e)
                r = requests.Response()  # empty Response object used when raising ConnectionError
            else:
                if r.status_code == 200:
                    if not json:
                        return r
                    elif not r.content:
                        msg = "GET: empty response"
                    else:
                        try:
                            return r.json()
                        except ValueError:
                            msg = f"GET: invalid JSON : {r.content!r}"
                else:
                    msg = f"GET: error {r.status_code}: {r.content!r}"
            msg = f"{msg}, {attempts_left} attempts left"
            if attempts_left <= 0:
                bioblend.log.error(msg)
                raise ConnectionError(
                    msg,
                    body=r.text,
                    status_code=r.status_code,
                )
            else:
                bioblend.log.warning(msg)
                time.sleep(retry_delay)

    def _post(
        self,
        payload: Optional[dict] = None,
        id: Optional[str] = None,
        deleted: bool = False,
        contents: bool = False,
        url: Optional[str] = None,
        files_attached: bool = False,
    ) -> Any:
        """
        Do a generic POST request, composing the url from the contents of the
        arguments. Alternatively, an explicit ``url`` can be provided to use
        for the request.
        The payload dict may contain file handles (in which case the
        ``files_attached`` flag must be set to true).

        If ``files_attached`` is set to ``False``, the request body will be
        JSON-encoded; otherwise, it will be encoded as multipart/form-data.

        :type payload: dict
        :param payload: additional parameters to send in the body of the request

        :return: The decoded response.
        """
        if not url:
            url = self._make_url(module_id=id, deleted=deleted, contents=contents)
        return self.gi.make_post_request(url, payload=payload, files_attached=files_attached)

    def _put(
        self,
        payload: Optional[dict] = None,
        id: Optional[str] = None,
        url: Optional[str] = None,
        params: Optional[dict] = None,
    ) -> Any:
        """
        Do a generic PUT request, composing the url from the contents of the
        arguments. Alternatively, an explicit ``url`` can be provided to use
        for the request.

        :type payload: dict
        :param payload: additional parameters to send in the body of the request

        :return: The decoded response.
        """
        if not url:
            url = self._make_url(module_id=id)
        return self.gi.make_put_request(url, payload=payload, params=params)

    def _patch(
        self,
        payload: Optional[dict] = None,
        id: Optional[str] = None,
        url: Optional[str] = None,
        params: Optional[dict] = None,
    ) -> Any:
        """
        Do a generic PATCH request, composing the url from the contents of the
        arguments. Alternatively, an explicit ``url`` can be provided to use
        for the request.

        :type payload: dict
        :param payload: additional parameters to send in the body of the request

        :return: The decoded response.
        """
        if not url:
            url = self._make_url(module_id=id)
        return self.gi.make_patch_request(url, payload=payload, params=params)

    def _delete(
        self,
        payload: Optional[dict] = None,
        id: Optional[str] = None,
        deleted: bool = False,
        contents: bool = False,
        url: Optional[str] = None,
        params: Optional[dict] = None,
    ) -> Any:
        """
        Do a generic DELETE request, composing the url from the contents of the
        arguments. Alternatively, an explicit ``url`` can be provided to use
        for the request.

        :type payload: dict
        :param payload: additional parameters to send in the body of the request

        :return: The decoded response or None.
        """
        if not url:
            url = self._make_url(module_id=id, deleted=deleted, contents=contents)
        r = self.gi.make_delete_request(url, payload=payload, params=params)
        if 200 <= r.status_code < 203:
            return r.json()
        elif 203 <= r.status_code < 300:
            return None
        # @see self.body for HTTP response body
        raise ConnectionError(
            f"Unexpected HTTP status code: {r.status_code}",
            body=r.text,
            status_code=r.status_code,
        )



"""
Contains possible interactions with the Galaxy library folders
"""

from typing import (
    Any,
    Dict,
    List,
    Literal,
    Optional,
    TYPE_CHECKING,
    Union,
)

from bioblend.galaxy.client import Client

if TYPE_CHECKING:
    from bioblend.galaxy import GalaxyInstance


class FoldersClient(Client):
    module = "folders"

    def __init__(self, galaxy_instance: "GalaxyInstance") -> None:
        super().__init__(galaxy_instance)

    def create_folder(self, parent_folder_id: str, name: str, description: Optional[str] = None) -> Dict[str, Any]:
        """
        Create a folder.

        :type parent_folder_id: str
        :param parent_folder_id: Folder's description

        :type name: str
        :param name: name of the new folder

        :type description: str
        :param description: folder's description

        :rtype: dict
        :return: details of the updated folder
        """
        payload: Dict[str, str] = {"name": name}
        if description:
            payload["description"] = description
        return self._post(payload=payload, id=parent_folder_id)

    def show_folder(self, folder_id: str, contents: bool = False) -> Dict[str, Any]:
        """
        Display information about a folder.

        :type folder_id: str
        :param folder_id: the folder's encoded id, prefixed by 'F'

        :type contents: bool
        :param contents: True to get the contents of the folder, rather
          than just the folder details.

        :rtype: dict
        :return: dictionary including details of the folder
        """

        return self._get(id=folder_id, contents=contents)

    def delete_folder(self, folder_id: str, undelete: bool = False) -> Dict[str, Any]:
        """
        Marks the folder with the given ``id`` as `deleted` (or removes the
        `deleted` mark if the `undelete` param is True).

        :type folder_id: str
        :param folder_id: the folder's encoded id, prefixed by 'F'

        :type undelete: bool
        :param undelete: If set to True, the folder will be undeleted
                         (i.e. the `deleted` mark will be removed)

        :return: detailed folder information
        :rtype: dict
        """
        payload = {"undelete": undelete}
        return self._delete(payload=payload, id=folder_id)

    def update_folder(self, folder_id: str, name: str, description: Optional[str] = None) -> Dict[str, Any]:
        """
        Update folder information.

        :type folder_id: str
        :param folder_id: the folder's encoded id, prefixed by 'F'

        :type name: str
        :param name: name of the new folder

        :type description: str
        :param description: folder's description

        :rtype: dict
        :return: details of the updated folder
        """
        payload = {"name": name}
        if description:
            payload["description"] = description
        return self._put(payload=payload, id=folder_id)

    def get_permissions(self, folder_id: str, scope: Literal["current", "available"] = "current") -> Dict[str, Any]:
        """
        Get the permissions of a folder.

        :type folder_id: str
        :param folder_id: the folder's encoded id, prefixed by 'F'

        :type scope: str
        :param scope: scope of permissions, either 'current' or 'available'

        :rtype: dict
        :return: dictionary including details of the folder permissions
        """
        url = self._make_url(folder_id) + "/permissions"
        return self._get(url=url)

    def set_permissions(
        self,
        folder_id: str,
        action: Literal["set_permissions"] = "set_permissions",
        add_ids: Optional[List[str]] = None,
        manage_ids: Optional[List[str]] = None,
        modify_ids: Optional[List[str]] = None,
    ) -> Dict[str, Any]:
        """
        Set the permissions of a folder.

        :type folder_id: str
        :param folder_id: the folder's encoded id, prefixed by 'F'

        :type action: str
        :param action: action to execute, only "set_permissions" is supported.

        :type add_ids: list of str
        :param add_ids: list of role IDs which can add datasets to the folder

        :type manage_ids: list of str
        :param manage_ids: list of role IDs which can manage datasets in the folder

        :type modify_ids: list of str
        :param modify_ids: list of role IDs which can modify datasets in the folder

        :rtype: dict
        :return: dictionary including details of the folder
        """
        url = self._make_url(folder_id) + "/permissions"
        payload: Dict[str, Union[str, List[str]]] = {"action": action}
        if add_ids:
            payload["add_ids[]"] = add_ids
        if manage_ids:
            payload["manage_ids[]"] = manage_ids
        if modify_ids:
            payload["modify_ids[]"] = modify_ids
        return self._post(url=url, payload=payload)



"""
A representation of a Galaxy instance based on oo wrappers.
"""

import time
from typing import (
    Iterable,
    List,
    Optional,
)

import bioblend
import bioblend.galaxy
from bioblend.galaxy.datasets import TERMINAL_STATES
from . import (
    client,
    wrappers,
)


def _get_error_info(dataset: wrappers.Dataset) -> str:
    msg = dataset.id
    try:
        msg += f" ({dataset.name}): {dataset.misc_info}"
    except Exception:  # avoid 'error while generating an error report'
        msg += ": error"
    return msg


class GalaxyInstance:
    """
    A representation of an instance of Galaxy, identified by a URL and
    a user's API key.

    :type url: str
    :param url: a FQDN or IP for a given instance of Galaxy. For example:
      ``http://127.0.0.1:8080``

    :type api_key: str
    :param api_key: user's API key for the given instance of Galaxy, obtained
      from the Galaxy web UI.

    This is actually a factory class which instantiates the entity-specific
    clients.

    Example: get a list of all histories for a user with API key 'foo'::

      from bioblend.galaxy.objects import GalaxyInstance
      gi = GalaxyInstance('http://127.0.0.1:8080', api_key='foo')
      histories = gi.histories.list()
    """

    def __init__(
        self,
        url: str,
        api_key: Optional[str] = None,
        email: Optional[str] = None,
        password: Optional[str] = None,
        *,
        verify: bool = True,
    ) -> None:
        self.gi = bioblend.galaxy.GalaxyInstance(url, key=api_key, email=email, password=password, verify=verify)
        self.log = bioblend.log
        self.datasets = client.ObjDatasetClient(self)
        self.dataset_collections = client.ObjDatasetCollectionClient(self)
        self.histories = client.ObjHistoryClient(self)
        self.libraries = client.ObjLibraryClient(self)
        self.workflows = client.ObjWorkflowClient(self)
        self.invocations = client.ObjInvocationClient(self)
        self.tools = client.ObjToolClient(self)
        self.jobs = client.ObjJobClient(self)

    def _wait_datasets(
        self, datasets: Iterable[wrappers.Dataset], polling_interval: float, break_on_error: bool = True
    ) -> None:
        """
        Wait for datasets to come out of the pending states.

        :type datasets: :class:`~collections.Iterable` of
          :class:`~.wrappers.Dataset`
        :param datasets: datasets

        :type polling_interval: float
        :param polling_interval: polling interval in seconds

        :type break_on_error: bool
        :param break_on_error: if ``True``, raise a RuntimeError exception as
          soon as at least one of the datasets is in the 'error' state.

        .. warning::

          This is a blocking operation that can take a very long time.
          Also, note that this method does not return anything;
          however, each input dataset is refreshed (possibly multiple
          times) during the execution.
        """

        def poll(ds_list: Iterable[wrappers.Dataset]) -> List[wrappers.Dataset]:
            pending = []
            for ds in ds_list:
                ds.refresh()
                if break_on_error and ds.state == "error":
                    raise RuntimeError(_get_error_info(ds))
                if not ds.state:
                    self.log.warning("Dataset %s has an empty state", ds.id)
                elif ds.state not in TERMINAL_STATES:
                    self.log.info("Dataset %s is in non-terminal state %s", ds.id, ds.state)
                    pending.append(ds)
            return pending

        self.log.info("Waiting for datasets")
        while datasets:
            datasets = poll(datasets)
            time.sleep(polling_interval)



"""
Interaction with a Tool Shed instance repositories
"""

from typing import (
    Any,
    Dict,
    List,
    Literal,
    Optional,
    TYPE_CHECKING,
    Union,
)

from bioblend.galaxy.client import Client
from bioblend.util import attach_file

if TYPE_CHECKING:
    from bioblend.toolshed import ToolShedInstance


class ToolShedRepositoryClient(Client):
    module = "repositories"

    def __init__(self, toolshed_instance: "ToolShedInstance") -> None:
        super().__init__(toolshed_instance)

    def get_repositories(self, name: Optional[str] = None, owner: Optional[str] = None) -> List[Dict[str, Any]]:
        """
        Get all repositories in a Galaxy Tool Shed, or select a subset by
        specifying optional arguments for filtering (e.g. a repository name).

        :type name: str
        :param name: Repository name to filter on.

        :type owner: str
        :param owner: Repository owner to filter on.

        :rtype: list
        :return: Returns a list of dictionaries containing information about
          repositories present in the Tool Shed.
          For example::

            [{'category_ids': ['c1df3132f6334b0e', 'f6d7b0037d901d9b'],
              'create_time': '2020-02-09T16:24:37.098176',
              'deleted': False,
              'deprecated': False,
              'description': 'Order Contigs',
              'homepage_url': '',
              'id': '287bd69f724b99ce',
              'model_class': 'Repository',
              'name': 'best_tool_ever',
              'owner': 'billybob',
              'private': False,
              'remote_repository_url': '',
              'times_downloaded': 0,
              'type': 'unrestricted',
              'user_id': '5cefd48bc04af6d4'}]

        .. versionchanged:: 0.4.1
          Changed method name from ``get_tools`` to ``get_repositories`` to
          better align with the Tool Shed concepts.
        """
        params = {}
        if name:
            params["name"] = name
        if owner:
            params["owner"] = owner
        return self._get(params=params)

    def search_repositories(
        self,
        q: str,
        page: int = 1,
        page_size: int = 10,
    ) -> Dict[str, Any]:
        """
        Search for repositories in a Galaxy Tool Shed.

        :type  q: str
        :param q: query string for searching purposes

        :type  page: int
        :param page: page requested

        :type  page_size: int
        :param page_size: page size requested

        :rtype:  dict
        :return: dictionary containing search hits as well as metadata for the
          search.
          For example::

            {'hits': [{'matched_terms': [],
                       'repository': {'approved': 'no',
                                      'categories': 'fastq manipulation',
                                      'description': 'Convert export file to fastq',
                                      'full_last_updated': '2015-01-18 09:48 AM',
                                      'homepage_url': '',
                                      'id': 'bdfa208f0cf6504e',
                                      'last_updated': 'less than a year',
                                      'long_description': 'This is a simple too to convert Solexas Export files to FASTQ files.',
                                      'name': 'export_to_fastq',
                                      'remote_repository_url': '',
                                      'repo_lineage': "['0:c9e926d9d87e', '1:38859774da87']"
                                      'repo_owner_username': 'louise',
                                      'times_downloaded': 164},
                       'score': 4.92},
                      {'matched_terms': [],
                       'repository': {'approved': 'no',
                                      'categories': 'fastq manipulation',
                                      'description': 'Convert BAM file to fastq',
                                      'full_last_updated': '2015-04-07 11:57 AM',
                                      'homepage_url': '',
                                      'id': '175812cd7caaf439',
                                      'last_updated': 'less than a month',
                                      'long_description': 'Use Picards SamToFastq to convert a BAM file to fastq. Useful for storing reads as BAM in Galaxy and converting to fastq when needed for analysis.',
                                      'name': 'bam_to_fastq',
                                      'remote_repository_url': '',
                                      'repo_lineage': "['0:a0af255e28c1', '1:2523cb0fb84c', '2:2656247b5253']"
                                      'repo_owner_username': 'brad-chapman',
                                      'times_downloaded': 138},
                       'score': 4.14}],
             'hostname': 'https://testtoolshed.g2.bx.psu.edu/',
             'page': '1',
             'page_size': '2',
             'total_results': '64'}
        """
        params = {"q": q, "page": page, "page_size": page_size}
        return self._get(params=params)

    def show_repository(
        self,
        toolShed_id: str,
    ) -> Dict[str, Any]:
        """
        Display information of a repository from Tool Shed

        :type toolShed_id: str
        :param toolShed_id: Encoded Tool Shed ID

        :rtype: dict
        :return: Information about the tool.
          For example::

            {'category_ids': ['c1df3132f6334b0e', 'f6d7b0037d901d9b'],
             'create_time': '2020-02-22T20:39:15.548491',
             'deleted': False,
             'deprecated': False,
             'description': 'Order Contigs',
             'homepage_url': '',
             'id': '287bd69f724b99ce',
             'long_description': '',
             'model_class': 'Repository',
             'name': 'best_tool_ever',
             'owner': 'billybob',
             'private': False,
             'remote_repository_url': '',
             'times_downloaded': 0,
             'type': 'unrestricted',
             'user_id': '5cefd48bc04af6d4'}

        .. versionchanged:: 0.4.1
          Changed method name from ``show_tool`` to ``show_repository`` to
          better align with the Tool Shed concepts.
        """
        return self._get(id=toolShed_id)

    def get_ordered_installable_revisions(
        self,
        name: str,
        owner: str,
    ) -> List[str]:
        """
        Returns the ordered list of changeset revision hash strings that are
        associated with installable revisions. As in the changelog, the list is
        ordered oldest to newest.

        :type name: str
        :param name: the name of the repository

        :type owner: str
        :param owner: the owner of the repository

        :rtype: list
        :return: List of changeset revision hash strings from oldest to newest
        """
        url = self._make_url() + "/get_ordered_installable_revisions"
        params = {"name": name, "owner": owner}
        r = self._get(url=url, params=params)

        return r

    def get_repository_revision_install_info(
        self,
        name: str,
        owner: str,
        changeset_revision: str,
    ) -> List[Dict[str, Any]]:
        """
        Return a list of dictionaries of metadata about a certain changeset
        revision for a single tool.

        :type name: str
        :param name: the name of the repository

        :type owner: str
        :param owner: the owner of the repository

        :type changeset_revision: str
        :param changeset_revision: the changeset_revision of the
          RepositoryMetadata object associated with the repository

        :rtype: List of dictionaries
        :return: Returns a list of the following dictionaries:

          #. a dictionary defining the repository
          #. a dictionary defining the repository revision (RepositoryMetadata)
          #. a dictionary including the additional information required to
             install the repository

          For example::

            [{'create_time': '2020-08-20T13:17:08.818518',
              'deleted': False,
              'deprecated': False,
              'description': 'Galaxy Freebayes Bayesian genetic variant detector tool',
              'homepage_url': '',
              'id': '491b7a3fddf9366f',
              'long_description': 'Galaxy Freebayes Bayesian genetic variant detector tool originally included in the Galaxy code distribution but migrated to the tool shed.',
              'model_class': 'Repository',
              'name': 'freebayes',
              'owner': 'devteam',
              'private': False,
              'remote_repository_url': '',
              'times_downloaded': 269,
              'type': 'unrestricted',
              'url': '/api/repositories/491b7a3fddf9366f',
              'user_id': '1de29d50c3c44272'},
             {'changeset_revision': 'd291dc763c4c',
              'do_not_test': False,
              'downloadable': True,
              'has_repository_dependencies': False,
              'id': '504be8aaa652c154',
              'includes_datatypes': False,
              'includes_tool_dependencies': True,
              'includes_tools': True,
              'includes_tools_for_display_in_tool_panel': True,
              'includes_workflows': False,
              'malicious': False,
              'missing_test_components': False,
              'model_class': 'RepositoryMetadata',
              'numeric_revision': 0,
              'repository_id': '491b7a3fddf9366f',
              'url': '/api/repository_revisions/504be8aaa652c154'},
              'valid_tools': [{'add_to_tool_panel': True,
                'description': '- Bayesian genetic variant detector',
                'guid': 'testtoolshed.g2.bx.psu.edu/repos/devteam/freebayes/freebayes/0.0.3',
                'id': 'freebayes',
                'name': 'FreeBayes',
                'requirements': [{'name': 'freebayes',
                  'type': 'package',
                  'version': '0.9.6_9608597d12e127c847ae03aa03440ab63992fedf'},
                {'name': 'samtools', 'type': 'package', 'version': '0.1.18'}],
                'tests': [{'inputs': [['reference_source|reference_source_selector',
                    'history'],
                  ['options_type|options_type_selector', 'basic'],
                  ['reference_source|ref_file', 'phiX.fasta'],
                  ['reference_source|input_bams_0|input_bam', 'fake_phiX_reads_1.bam']],
                  'name': 'Test-1',
                  'outputs': [['output_vcf', 'freebayes_out_1.vcf.contains']],
                  'required_files': ['fake_phiX_reads_1.bam',
                  'phiX.fasta',
                  'freebayes_out_1.vcf.contains']}],
                'tool_config': '/srv/toolshed/test/var/data/repos/000/repo_708/freebayes.xml',
                'tool_type': 'default',
                'version': '0.0.3',
                'version_string_cmd': None}]},
             {'freebayes': ['Galaxy Freebayes Bayesian genetic variant detector tool',
                            'http://testtoolshed.g2.bx.psu.edu/repos/devteam/freebayes',
                            'd291dc763c4c',
                            '9',
                            'devteam',
                            {},
                            {'freebayes/0.9.6_9608597d12e127c847ae03aa03440ab63992fedf': {'changeset_revision': 'd291dc763c4c',
                                                                                          'name': 'freebayes',
                                                                                          'repository_name': 'freebayes',
                                                                                          'repository_owner': 'devteam',
                                                                                          'type': 'package',
                                                                                          'version': '0.9.6_9608597d12e127c847ae03aa03440ab63992fedf'},
                             'samtools/0.1.18': {'changeset_revision': 'd291dc763c4c',
                                                 'name': 'samtools',
                                                 'repository_name': 'freebayes',
                                                 'repository_owner': 'devteam',
                                                 'type': 'package',
                                                 'version': '0.1.18'}}]}]
        """
        url = self._make_url() + "/get_repository_revision_install_info"
        params = {"name": name, "owner": owner, "changeset_revision": changeset_revision}
        return self._get(url=url, params=params)

    def repository_revisions(
        self,
        downloadable: Optional[bool] = None,
        malicious: Optional[bool] = None,
        missing_test_components: Optional[bool] = None,
        includes_tools: Optional[bool] = None,
    ) -> List[Dict[str, Any]]:
        """
        Returns a (possibly filtered) list of dictionaries that include
        information about all repository revisions. The following parameters can
        be used to filter the list.

        :type downloadable: bool
        :param downloadable: Can the tool be downloaded

        :type malicious: bool
        :param malicious:

        :type missing_test_components: bool
        :param missing_test_components:

        :type includes_tools: bool
        :param includes_tools:

        :rtype: List of dictionaries
        :return: Returns a (possibly filtered) list of dictionaries that include
          information about all repository revisions.
          For example::

            [{'changeset_revision': '6e26c5a48e9a',
              'downloadable': True,
              'has_repository_dependencies': False,
              'id': '92250afff777a169',
              'includes_datatypes': False,
              'includes_tool_dependencies': False,
              'includes_tools': True,
              'includes_tools_for_display_in_tool_panel': True,
              'includes_workflows': False,
              'malicious': False,
              'missing_test_components': False,
              'model_class': 'RepositoryMetadata',
              'numeric_revision': None,
              'repository_id': '78f2604ff5e65707',
              'url': '/api/repository_revisions/92250afff777a169'},
             {'changeset_revision': '15a54fa11ad7',
              'downloadable': True,
              'has_repository_dependencies': False,
              'id': 'd3823c748ae2205d',
              'includes_datatypes': False,
              'includes_tool_dependencies': False,
              'includes_tools': True,
              'includes_tools_for_display_in_tool_panel': True,
              'includes_workflows': False,
              'malicious': False,
              'missing_test_components': False,
              'model_class': 'RepositoryMetadata',
              'numeric_revision': None,
              'repository_id': 'f9662009da7bfce0',
              'url': '/api/repository_revisions/d3823c748ae2205d'}]
        """
        # Not using '_make_url' or '_get' to create url since the module id used
        # to create url is not the same as needed for this method
        url = self.gi.url + "/repository_revisions"
        params = {}
        if downloadable is not None:
            params["downloadable"] = downloadable
        if malicious is not None:
            params["malicious"] = malicious
        if missing_test_components is not None:
            params["missing_test_components"] = missing_test_components
        if includes_tools is not None:
            params["includes_tools"] = includes_tools
        return self._get(url=url, params=params)

    def show_repository_revision(
        self,
        metadata_id: str,
    ) -> Dict[str, Any]:
        """
        Returns a dictionary that includes information about a specified
        repository revision.

        :type metadata_id: str
        :param metadata_id: Encoded repository metadata ID

        :rtype: dict
        :return: Returns a dictionary that includes information about a
          specified repository revision.
          For example::

            {'changeset_revision': '7602de1e7f32',
             'downloadable': True,
             'has_repository_dependencies': False,
             'id': '504be8aaa652c154',
             'includes_datatypes': False,
             'includes_tool_dependencies': False,
             'includes_tools': True,
             'includes_tools_for_display_in_tool_panel': True,
             'includes_workflows': False,
             'malicious': False,
             'missing_test_components': True,
             'model_class': 'RepositoryMetadata',
             'numeric_revision': None,
             'repository_dependencies': [],
             'repository_id': '491b7a3fddf9366f',
             'url': '/api/repository_revisions/504be8aaa652c154'}
        """
        # Not using '_make_url' or '_get' to create url since the module id used
        # to create url is not the same as needed for this method
        # since metadata_id has to be defined, easy to create the url here
        url = "/".join((self.gi.url, "repository_revisions", metadata_id))
        return self._get(url=url)

    def update_repository(self, id: str, tar_ball_path: str, commit_message: Optional[str] = None) -> Dict[str, Any]:
        """
        Update the contents of a Tool Shed repository with specified tar ball.

        :type id: str
        :param id: Encoded repository ID

        :type tar_ball_path: str
        :param tar_ball_path: Path to file containing tar ball to upload.

        :type commit_message: str
        :param commit_message: Commit message used for the underlying Mercurial
          repository backing Tool Shed repository.

        :rtype: dict
        :return: Returns a dictionary that includes repository content warnings.
          Most valid uploads will result in no such warning and an exception
          will be raised generally if there are problems.
          For example a successful upload will look like::

            {'content_alert': '',
             'message': ''}

        .. versionadded:: 0.5.2
        """
        url = self._make_url(id) + "/changeset_revision"
        payload: Dict[str, Any] = {"file": attach_file(tar_ball_path)}
        if commit_message is not None:
            payload["commit_message"] = commit_message
        try:
            return self._post(payload=payload, files_attached=True, url=url)
        finally:
            payload["file"].close()

    def create_repository(
        self,
        name: str,
        synopsis: str,
        description: Optional[str] = None,
        type: Literal["unrestricted", "repository_suite_definition", "tool_dependency_definition"] = "unrestricted",
        remote_repository_url: Optional[str] = None,
        homepage_url: Optional[str] = None,
        category_ids: Optional[List[str]] = None,
    ) -> Dict[str, Any]:
        """
        Create a new repository in a Tool Shed.

        :type name: str
        :param name: Name of the repository

        :type synopsis: str
        :param synopsis: Synopsis of the repository

        :type description: str
        :param description: Optional description of the repository

        :type type: str
        :param type: type of the repository. One of "unrestricted",
          "repository_suite_definition", or "tool_dependency_definition"

        :type remote_repository_url: str
        :param remote_repository_url: Remote URL (e.g. GitHub/Bitbucket
          repository)

        :type homepage_url: str
        :param homepage_url: Upstream's homepage for the project

        :type category_ids: list
        :param category_ids: List of encoded category IDs

        :rtype: dict
        :return: a dictionary containing information about the new repository.
          For example::

            {"deleted": false,
             "deprecated": false,
             "description": "new_synopsis",
             "homepage_url": "https://github.com/galaxyproject/",
             "id": "8cf91205f2f737f4",
             "long_description": "this is some repository",
             "model_class": "Repository",
             "name": "new_repo_17",
             "owner": "qqqqqq",
             "private": false,
             "remote_repository_url": "https://github.com/galaxyproject/tools-devteam",
             "times_downloaded": 0,
             "type": "unrestricted",
             "user_id": "adb5f5c93f827949"}
        """
        payload: Dict[str, Union[str, List[str]]] = {
            "name": name,
            "synopsis": synopsis,
        }
        if description is not None:
            payload["description"] = description
        if description is not None:
            payload["description"] = description
        if type is not None:
            payload["type"] = type
        if remote_repository_url is not None:
            payload["remote_repository_url"] = remote_repository_url
        if homepage_url is not None:
            payload["homepage_url"] = homepage_url
        if category_ids is not None:
            payload["category_ids[]"] = category_ids
        return self._post(payload)

    def update_repository_metadata(
        self,
        toolShed_id: str,
        name: Optional[str] = None,
        synopsis: Optional[str] = None,
        description: Optional[str] = None,
        remote_repository_url: Optional[str] = None,
        homepage_url: Optional[str] = None,
        category_ids: Optional[List[str]] = None,
    ) -> Dict[str, Any]:
        """
        Update metadata of a Tool Shed repository.

        :type toolShed_id: str
        :param name: ID of the repository to update

        :type name: str
        :param name: New name of the repository

        :type synopsis: str
        :param synopsis: New synopsis of the repository

        :type description: str
        :param description: New description of the repository

        :type remote_repository_url: str
        :param remote_repository_url: New remote URL (e.g. GitHub/Bitbucket
          repository)

        :type homepage_url: str
        :param homepage_url: New upstream homepage for the project

        :type category_ids: list
        :param category_ids: New list of encoded category IDs

        :rtype: dict
        :return: a dictionary containing information about the updated repository.
        """
        payload: Dict[str, Union[str, List[str]]] = {}
        if name:
            payload["name"] = name
        if synopsis:
            payload["synopsis"] = synopsis
        if description:
            payload["description"] = description
        if remote_repository_url:
            payload["remote_repository_url"] = remote_repository_url
        if homepage_url:
            payload["homepage_url"] = homepage_url
        if category_ids:
            payload["category_ids"] = category_ids
        return self._put(id=toolShed_id, payload=payload)



"""
Contains possible interactions with the Galaxy visualization
"""

from typing import (
    Any,
    Dict,
    List,
    TYPE_CHECKING,
)

from bioblend.galaxy.client import Client

if TYPE_CHECKING:
    from bioblend.galaxy import GalaxyInstance


class VisualClient(Client):
    module = "visualizations"

    def __init__(self, galaxy_instance: "GalaxyInstance") -> None:
        super().__init__(galaxy_instance)

    def get_visualizations(self) -> List[Dict[str, Any]]:
        """
        Get the list of all visualizations.

        :rtype: list
        :return: A list of dicts with details on individual visualizations.
          For example::

            [{'dbkey': 'eschColi_K12',
              'id': 'df1c7c96fc427c2d',
              'title': 'AVTest1',
              'type': 'trackster',
              'url': '/api/visualizations/df1c7c96fc427c2d'},
             {'dbkey': 'mm9',
              'id': 'a669f50f8bf55b02',
              'title': 'Bam to Bigwig',
              'type': 'trackster',
              'url': '/api/visualizations/a669f50f8bf55b02'}]
        """
        return self._get()

    def show_visualization(self, visual_id: str) -> Dict[str, Any]:
        """
        Get details of a given visualization.

        :type visual_id: str
        :param visual_id: Encoded visualization ID

        :rtype: dict
        :return: A description of the given visualization.
          For example::

            {'annotation': None,
             'dbkey': 'mm9',
             'id': '18df9134ea75e49c',
             'latest_revision': {  ... },
             'model_class': 'Visualization',
             'revisions': ['aa90649bb3ec7dcb', '20622bc6249c0c71'],
             'slug': 'visualization-for-grant-1',
             'title': 'Visualization For Grant',
             'type': 'trackster',
             'url': '/u/azaron/v/visualization-for-grant-1',
             'user_id': '21e4aed91386ca8b'}
        """
        return self._get(id=visual_id)
